#!/usr/bin/env python3.9
import datetime
from importlib import resources  # only works from Python 3.7
import sys
from typing import Dict, IO, List
import logging
import lark
import objmodel  # other ways of importing do not allow updating `top`
import terminals
import other
import instantiator
from bodyprocessor import BodyProcessor
from tools import larkify


def print_with_imports(modNames: List[str],
                       modDict: Dict[str, objmodel.MaudeElem],
                       f: IO,
                       done: List[str]) -> None:
    # Outputs all modules in `modNames` and all their dependencies.
    for mn in modNames:
        m = modDict.get(mn)
        if m is not None and isinstance(m, objmodel.Module)\
                and m.name not in done:
            done.append(m.name)
            bp = BodyProcessor(m.body)
            imps = bp.protectings.copy()
            imps += bp.extendings
            imps += bp.includings
            for i in imps:
                print_with_imports([i.mod_expr.name], modDict, f, done)
            f.write(str(m) + "\n\n")


def mods_with_ag(mods: Dict[str, objmodel.MaudeElem]) -> List[str]:
    # Selects and returns the modules with some `ag` statement.
    # TODO: This can surely be done in a better way.
    res = []
    for n in mods.keys():
        m = mods[n]
        if "<ag" in str(m):
            res.append(n)
    return res


def main():
    log_file_name = "logs/" + str(datetime.datetime.now()) + ".log"
    logging.basicConfig(filename=log_file_name, level=logging.INFO)
    logging.info("Started at " + str(datetime.datetime.now()))
    if len(sys.argv) < 4:
        raise Exception('Usage: python3 main.py c|v infile module')
    else:
        oper = sys.argv[1]
        inFileName = sys.argv[2]
        modName = sys.argv[3]

    inFile = open(inFileName, "r")
    inCMaudeText = inFile.read()
    inFile.close()
    larkText = resources.read_text('grammars', 'cmaude.lark')
    cMaudeGrammarText = terminals.terminal_defs() + larkText
    cMaudeGrammar = lark.Lark(larkify(cMaudeGrammarText), start="maude_top")
    inCMaudeAst = cMaudeGrammar.parse(inCMaudeText)
    top = other.astToObjModel(inCMaudeAst)
    top.elems.insert(0, other.stageMod)

    instTop = instantiator.Instantiator.params_to_instances(top)
    mods: Dict[str, objmodel.MaudeElem] = {}
    if oper == 'c':
        for e in instTop.elems:
            mods[e.name] = other.make_std_comp_trans(instTop, e)
        outFileName = inFileName.rsplit('.', 1)[0] + '-c.maude'
        with open(outFileName, 'w') as of:
            print_with_imports([modName], mods, of, [])
    elif oper == 'v':
        for e in instTop.elems:
            mods[e.name] = other.make_std_for_verif(instTop, e)
        outFileName = inFileName.rsplit(".", 1)[0] + '-v.maude'
        modsToOutput = mods_with_ag(mods) + [modName]
        with open(outFileName, 'w') as of:
            print_with_imports(modsToOutput, mods, of, [])
    elif oper == 'p':
        for e in instTop.elems:
            mods[e.name] = other.make_std_partit_trans(instTop, e)
        outFileName = inFileName.rsplit('.', 1)[0] + '-p.maude'
        with open(outFileName, 'w') as of:
            print_with_imports([modName], mods, of, [])
    else:
        raise Exception('Usage: python3 main.py c|v infile module')
    logging.info("Ended at " + str(datetime.datetime.now()))


if __name__ == '__main__':
    main()
