#!/usr/bin/env python3.9


import itertools
from typing import Union, Tuple
from objmodel import *


class BodyProcessor:
    def __init__(self, b: ModBody) -> None:
        """The following code produces *normalized* lists of declarations for
        sorts, subsorts, ops, vars, and ppt. That is, even if the input
        module has `sort S1 S2 .`, this method produces `sort S1 .` and `sort
        S2 .`. This allows for an easier treatment in other parts of the
        code... though I'm not sure I'm taking advantage of it anywhere."""
        self.__instrs = b.instrs
        self.__protectings = []
        self.__extendings = []
        self.__includings = []
        self.__sort_decls = []
        self.__subsort_decls = []
        self.__op_decls = []
        self.__var_decls = []
        self.__eqs = []
        self.__cond_eqs = []
        self.__mbs = []
        self.__cond_mbs = []
        self.__rules = []
        self.__cond_rules = []
        self.__eg_rules = []
        self.__cond_eg_rules = []
        self.__ppt_decls = []
        self.__ag_stats = []
        self.__sync_on = None
        self.__inh_stats = []
        self.__comments = []
        for i in self.__instrs:
            if isinstance(i, Protecting):
                self.__protectings.append(i)
            elif isinstance(i, Extending):
                self.__extendings.append(i)
            elif isinstance(i, Including):
                self.__includings.append(i)
            elif isinstance(i, SortDecl):
                self.ids = i.sorts.ids
                for s in self.ids:
                    self.__sort_decls.append(SortDecl(IdList(s)))
            elif isinstance(i, SubsortDecl):
                sorts = [ill.ids for ill in i.body.sort_lists]
                for (p1, p2) in zip(sorts[:-1], sorts[1:]):
                    for s1 in p1:
                        for s2 in p2:
                            self.__subsort_decls.append(
                                SubsortDecl(SubsortBody(IdList(s1),
                                                        IdList(s2))))
            elif isinstance(i, OpDecl):
                for o in i.ops.ids:
                    self.__op_decls.append(OpDecl(IdList(o), i.from_sorts,
                                                  i.arrow, i.to_sort, i.attrs))
            elif isinstance(i, VarDecl):
                for v in i.ids.ids:
                    self.__var_decls.append(VarDecl(IdList(v), i.sort))
            elif isinstance(i, Eq):
                self.__eqs.append(i)
            elif isinstance(i, CondEq):
                self.__cond_eqs.append(i)
            elif isinstance(i, Mb):
                self.__mbs.append(i)
            elif isinstance(i, CondMb):
                self.__cond_mbs.append(i)
            elif isinstance(i, EgRule):
                self.__eg_rules.append(i)
            elif isinstance(i, CondEgRule):
                self.__cond_eg_rules.append(i)
            elif isinstance(i, Rule):
                self.__rules.append(i)
            elif isinstance(i, CondRule):
                self.__cond_rules.append(i)
            elif isinstance(i, PptDecl):
                for p in i.ppts.ids:
                    self.__ppt_decls.append(PptDecl(IdList(p), i.from_sorts,
                                                    i.arrow, i.to_sort))
            elif isinstance(i, AGStat):
                self.__ag_stats.append(i)
            elif isinstance(i, SyncOn):
                self.__sync_on = i
            elif isinstance(i, InhStat):
                self.__inh_stats.append(i)

    @property
    def protectings(self) -> List[Protecting]:
        return self.__protectings

    @property
    def extendings(self) -> List[Extending]:
        return self.__extendings

    @property
    def includings(self) -> List[Including]:
        return self.__includings

    @property
    def sort_decls(self) -> List[SortDecl]:
        return self.__sort_decls

    @property
    def subsort_decls(self) -> List[SubsortDecl]:
        return self.__subsort_decls

    @property
    def op_decls(self) -> List[OpDecl]:
        return self.__op_decls

    @property
    def var_decls(self) -> List[VarDecl]:
        return self.__var_decls

    @property
    def eqs(self) -> List[Eq]:
        return self.__eqs

    @property
    def cond_eqs(self) -> List[CondEq]:
        return self.__cond_eqs

    @property
    def mbs(self) -> List[Mb]:
        return self.__mbs

    @property
    def cond_mbs(self) -> List[CondMb]:
        return self.__cond_mbs

    @property
    def rules(self) -> List[Rule]:
        return self.__rules

    @property
    def cond_rules(self) -> List[CondRule]:
        return self.__cond_rules

    @property
    def eg_rules(self) -> List[EgRule]:
        return self.__eg_rules

    @property
    def cond_eg_rules(self) -> List[CondEgRule]:
        return self.__cond_eg_rules

    @property
    def ppt_decls(self) -> List[PptDecl]:
        return self.__ppt_decls

    @property
    def ag_stats(self) -> List[AGStat]:
        return self.__ag_stats

    @property
    def sync_on(self) -> Union[SyncOn, None]:
        return self.__sync_on

    @property
    def inh_stats(self) -> List[InhStat]:
        return self.__inh_stats

    @property
    def all_imports(self) -> List[Instruction]:
        res = []
        res += self.protectings
        res += self.extendings
        res += self.includings
        return res

    @property
    def all_decls(self) -> List[Instruction]:
        res = []
        res += self.sort_decls
        res += self.subsort_decls
        res += self.op_decls
        res += self.ppt_decls
        res += self.var_decls
        return res

    @property
    def all_non_eg_rules(self) -> List[Union[Rule, CondRule]]:
        res = []
        res += self.rules
        res += self.cond_rules
        return res

    @property
    def all_eg_rules(self) -> List[Union[EgRule, CondEgRule]]:
        res = []
        res += self.eg_rules
        res += self.cond_eg_rules
        return res

    def get_declared_sorts(self) -> List[str]:
        return [s for sd in self.sort_decls for s in sd.sorts.ids]

    def get_subsort_items(self) -> List[Tuple[str, str]]:
        res = []
        for ssd in self.subsort_decls:
            ill = [sd.ids for sd in ssd.body.sort_lists]
            for (l1, l2) in zip(ill[:-1], ill[1:]):
                res += itertools.product(l1, l2)
        return res

    def get_declared_names(self) -> List[str]:
        """This takes the signature part of `self` (that is, declarations)
        and produces a list of the names declared in it.
        """
        res = []
        res += self.get_declared_sorts()
        for od in self.op_decls:
            for on in od.ops.ids:
                if on != '_@_':
                    res.append(on)
        for vd in self.var_decls:
            res += vd.ids.ids
        for pd in self.ppt_decls:
            res += pd.ppts.ids
        return res
