#!/usr/bin/env python3.9


from functools import singledispatchmethod
from typing import Dict
import re
from objmodel import *
from bodyprocessor import BodyProcessor


class Inliner:
    def __init__(self, vars_dict: Dict[str, str]) -> None:
        self.__vars_dict = vars_dict

    @staticmethod
    def inline_vars_in_body(b: ModBody) -> ModBody:
        vars_dict = {vd: vds.sort for vds in BodyProcessor(b).var_decls
                     for vd in vds.ids.ids}
        # I assume only one var per decl (it holds the way var_decls is coded)
        inlr = Inliner(vars_dict)
        instrs = [inlr.__inline_vars(i) for i in b.instrs
                  if not isinstance(i, VarDecl)]
        return ModBody(*instrs)

    @singledispatchmethod
    def __inline_vars(self, arg):
        raise NotImplementedError(
            f"There is no var inlining method for {type(arg)}.")

    @__inline_vars.register
    def _(self, s: str) -> str:
        # s is supposed to be a stringified term
        res = s
        for vn in self.__vars_dict.keys():
            res = re.sub(rf"(?<![a-zA-Z|<']){re.escape(vn)}(?![a-zA-Z|>':])",
                         vn + ":" + self.__vars_dict[vn], res)
        return res

    @__inline_vars.register
    def _(self, ac: AtomicCondition) -> AtomicCondition:
        raise Exception("Abstract method: to be implemented in subclasses.")

    @__inline_vars.register
    def _(self, ec: EqualCondition) -> EqualCondition:
        return EqualCondition(Term(self.__inline_vars(str(ec.lhs))),
                              Term(self.__inline_vars(str(ec.rhs))))

    @__inline_vars.register
    def _(self, ac: AssignCondition) -> AssignCondition:
        return AssignCondition(Term(self.__inline_vars(str(ac.lhs))),
                               Term(self.__inline_vars(str(ac.rhs))))

    @__inline_vars.register
    def _(self, mc: MemberCondition) -> MemberCondition:
        return MemberCondition(Term(self.__inline_vars(str(mc.lhs))),
                               self.__inline_vars(mc.rhs))

    @__inline_vars.register
    def _(self, rc: RewCondition) -> RewCondition:
        return RewCondition(Term(self.__inline_vars(str(rc.lhs))),
                            Term(self.__inline_vars(str(rc.rhs))))

    @__inline_vars.register
    def _(self, bc: BoolCondition) -> BoolCondition:
        return BoolCondition(Term(self.__inline_vars(str(bc.text))))

    @__inline_vars.register
    def _(self, c: Condition) -> Condition:
        return Condition(*[self.__inline_vars(ac) for ac in c.conds])

    @__inline_vars.register
    def _(self, i: Instruction) -> Instruction:
        return i

    @__inline_vars.register
    def _(self, e: Eq) -> Eq:
        return Eq(Term(self.__inline_vars(str(e.lhs))),
                  Term(self.__inline_vars(str(e.rhs))), e.attrs)

    @__inline_vars.register
    def _(self, ce: CondEq) -> CondEq:
        return CondEq(Term(self.__inline_vars(str(ce.lhs))),
                      Term(self.__inline_vars(str(ce.rhs))),
                      self.__inline_vars(ce.cond), ce.attrs)

    @__inline_vars.register
    def _(self, m: Mb) -> Mb:
        return Mb(Term(self.__inline_vars(str(m.term))), m.sort)

    @__inline_vars.register
    def _(self, cm: CondMb) -> CondMb:
        return CondMb(Term(self.__inline_vars(str(cm.term))),
                      cm.sort, self.__inline_vars(cm.cond))

    @__inline_vars.register
    def _(self, r: Rule) -> Rule:
        return Rule(Term(self.__inline_vars(str(r.lhs))),
                    Term(self.__inline_vars(str(r.rhs))),
                    r.attrs)

    @__inline_vars.register
    def _(self, cr: CondRule) -> CondRule:
        return CondRule(Term(self.__inline_vars(str(cr.lhs))),
                        Term(self.__inline_vars(str(cr.rhs))),
                        self.__inline_vars(cr.cond), cr.attrs)

    @__inline_vars.register
    def _(self, er: EgRule) -> EgRule:
        return EgRule(Term(self.__inline_vars(str(er.lhs))),
                      Term(self.__inline_vars(str(er.mid))),
                      Term(self.__inline_vars(str(er.rhs))),
                      er.attrs)

    @__inline_vars.register
    def _(self, cer: CondEgRule) -> CondEgRule:
        return CondEgRule(Term(self.__inline_vars(str(cer.lhs))),
                          Term(self.__inline_vars(str(cer.mid))),
                          Term(self.__inline_vars(str(cer.rhs))),
                          self.__inline_vars(cer.cond), cer.attrs)
