#!/usr/bin/env python3.9


from functools import singledispatchmethod
from typing import Dict
import re
from objmodel import *
from tools import buildName
from other import paramEmodImports, get_mod_by_name
from topmethods import get_view_by_name, get_ppt_mappings


class Instantiator:
    def __init__(self, modMap: Dict[str, str], pptMap: Dict[str, str]) -> None:
        self.__modMap = modMap
        self.__pptMap = pptMap

    @staticmethod
    def params_to_instances(top: MaudeTop) -> MaudeTop:
        """Looks into each element in `top`. Omits `emod`s with parameters and
        creates the needed instantiations, according to import instructions.
        """
        elts = []
        for e in top.elems:
            if isinstance(e, EModule) and len(e.params.params) > 0:
                pass
            elif isinstance(e, EModule) and len(paramEmodImports(e)) > 0:
                for i in paramEmodImports(e):
                    elts.append(Instantiator.__create_instance(i.mod_expr, top))
                elts.append(Instantiator.__with_inst_imports(e))
            else:
                elts.append(e)
        return MaudeTop(*elts)

    @staticmethod
    def __create_instance(me: ModExpr, top: MaudeTop) -> Module:
        """Creates the instantiation of the module given by `me`."""
        assert isinstance(me, ModExprWArgs)
        m = get_mod_by_name(top, me.name)
        formals = [pd.name for pd in m.params.params]
        views = me.args
        toModList = [get_view_by_name(top, vn).mod_to for vn in views.ids]
        pptNameMap = dict()
        for (viewName, parName, toModName)\
                in zip(views.ids, formals, toModList):
            view = get_view_by_name(top, viewName)
            for (lppt, rppt) in get_ppt_mappings(top, view):
                pptNameMap[parName + "$" + lppt] = toModName + "$" + rppt
        modNameMap = dict(zip([pd for pd in formals], toModList))
        instBody = Instantiator(modNameMap, pptNameMap).__instantiate(m.body)
        return EModule(buildName(me.name, *me.args.ids), ParamList(), instBody)

    @staticmethod
    def __with_inst_imports(em: EModule) -> EModule:
        """ Replaces import instructions for emods with argument by the import
        or the corresponding instantiated module, as created by
        `__create_instance` For example, `inc M{A, B} .` by ` inc <M|A|B> .`.
        """
        instrs = []
        for i in em.body.instrs:
            if (isinstance(i, Including) or isinstance(i, Extending)
                or isinstance(i, Protecting)) \
                    and isinstance(i.mod_expr, ModExprWArgs):
                instrs += [i.__class__(ModExprId(buildName(
                    i.mod_expr.name, *i.mod_expr.args.ids)))]
                # TODO: I don't like using __class__
            else:
                instrs.append(i)
        return EModule(em.name, em.params, ModBody(*instrs))

    @singledispatchmethod
    def __instantiate(self, arg):
        raise NotImplementedError(
            f"There is no instantiation method for {type(arg)}.")

    @__instantiate.register
    def _(self, s: str) -> str:
        res = s
        for p in self.__pptMap.keys():
            res = re.sub(rf"(?<![a-zA-Z]){re.escape(p)}(?![a-zA-Z])",
                         self.__pptMap[p], res)
        return res

    @__instantiate.register
    def _(self, ac: AtomicCondition) -> AtomicCondition:
        raise Exception("Method cannot be called for AtomicCondition,"
                        " only for some of its subclasses.")

    @__instantiate.register
    def _(self, ec: EqualCondition) -> EqualCondition:
        return EqualCondition(
            Term(self.__instantiate(str(ec.lhs))),
            Term(self.__instantiate(str(ec.rhs))))

    @__instantiate.register
    def _(self, ac: AssignCondition) -> AssignCondition:
        return AssignCondition(
            Term(self.__instantiate(str(ac.lhs))),
            Term(self.__instantiate(str(ac.rhs))))

    @__instantiate.register
    def _(self, mc: MemberCondition) -> MemberCondition:
        return MemberCondition(
            Term(self.__instantiate(str(mc.lhs))),
            self.__instantiate(mc.rhs))

    @__instantiate.register
    def _(self, rc: RewCondition) -> RewCondition:
        return RewCondition(Term(self.__instantiate(str(rc.lhs))),
                            Term(self.__instantiate(str(rc.rhs))))

    @__instantiate.register
    def _(self, bc: BoolCondition) -> BoolCondition:
        return BoolCondition(Term(self.__instantiate(str(bc.text))))

    @__instantiate.register
    def _(self, c: Condition) -> Condition:
        res = []
        for ac in c.conds:
            res.append(self.__instantiate(ac))
        return Condition(*res)

    @__instantiate.register
    def _(self, i: Instruction) -> Instruction:
        return i

    @__instantiate.register
    def _(self, e: Eq) -> Eq:
        return Eq(Term(self.__instantiate(str(e.lhs))),
                  Term(self.__instantiate(str(e.rhs))), e.attrs)

    @__instantiate.register
    def _(self, ce: CondEq) -> CondEq:
        return CondEq(Term(self.__instantiate(str(ce.lhs))),
                      Term(self.__instantiate(str(ce.rhs))),
                      self.__instantiate(ce.cond), ce.attrs)

    @__instantiate.register
    def _(self, so: SyncOn) -> SyncOn:
        mods = [self.__modMap[m] for m in so.mods.ids]
        crits = []
        for c in so.crits.crits:
            lc = c.lhs
            rc = c.rhs
            lc2 = self.__pptMap[lc]
            rc2 = self.__pptMap[rc]
            crits += [EqCrit(lc2, rc2) if isinstance(c, EqCrit)
                      else AssignCrit(lc2, rc2)]
        return SyncOn(ModList(*mods), CritList(*crits))

    @__instantiate.register
    def _(self, s: InhStat) -> InhStat:
        return InhStat(self.__instantiate(s.lhs),
                       self.__instantiate(s.rhs))

    @__instantiate.register
    def _(self, mb: ModBody) -> ModBody:
        return ModBody(*[self.__instantiate(i) for i in mb.instrs])
