from typing import List
from tools import indent


class Term:
    def __init__(self, *tokens: str) -> None:
        self.__tokens = [t.strip() for t in list(tokens)]

    @property
    def tokens(self) -> List[str]:
        return self.__tokens

    def __str__(self) -> str:
        return ' '.join(self.tokens)


class MaudeElem:
    @property
    def name(self) -> str:
        raise NotImplementedError("This should be implemented in subclasses.")


class MaudeTop:
    def __init__(self, *elems: MaudeElem) -> None:
        self.__elems = list(elems)

    @property
    def elems(self) -> List[MaudeElem]:
        return self.__elems

    def __str__(self) -> str:
        return "\n".join([str(e) for e in self.elems])


class ModExpr:
    @property
    def name(self) -> str:
        raise NotImplementedError("This should be implemented in subclasses.")


class ParamDecl:
    def __init__(self, name: str, mod_expr: ModExpr) -> None:
        self.__name = name.strip()
        self.__mod_expr = mod_expr

    @property
    def name(self) -> str:
        return self.__name

    @property
    def mod_expr(self) -> ModExpr:
        return self.__mod_expr

    def __str__(self) -> str:
        return self.name\
               + " " + "::"\
               + " " + str(self.mod_expr)


class ParamList:
    def __init__(self, *params: ParamDecl) -> None:
        self.__params = list(params)

    @property
    def params(self) -> List[ParamDecl]:
        return self.__params

    def __str__(self) -> str:
        if len(self.params) > 0:
            return "{"\
                   + ", ".join([str(x) for x in self.params])\
                   + "}"
        else:
            return ""


class Instruction:
    pass


class ModBody:
    def __init__(self, *instrs: Instruction) -> None:
        self.__instrs = list(instrs)

    @property
    def instrs(self) -> List[Instruction]:
        return self.__instrs

    def __str__(self) -> str:
        return "\n".join([str(x) for x in self.instrs])


class Module(MaudeElem):
    @property
    def name(self) -> str:
        raise NotImplementedError("This should be implemented in subclasses.")

    @property
    def params(self) -> ParamList:
        raise NotImplementedError("This should be implemented in subclasses.")

    @property
    def body(self) -> ModBody:
        raise NotImplementedError("This should be implemented in subclasses.")


class SModule(Module):
    def __init__(self, name: str, params: ParamList, body: ModBody) -> None:
        self.__name = name
        self.__params = params
        self.__body = body

    @property
    def name(self) -> str:
        return self.__name

    @property
    def params(self) -> ParamList:
        return self.__params

    @property
    def body(self) -> ModBody:
        return self.__body

    def __str__(self) -> str:
        return "mod"\
               + " " + self.name\
               + str(self.params)\
               + " " + "is"\
               + "\n" + indent(str(self.body))\
               + "\n" + "endm"


class FModule(Module):
    def __init__(self, name: str, params: ParamList, body: ModBody) -> None:
        self.__name = name
        self.__params = params
        self.__body = body

    @property
    def name(self) -> str:
        return self.__name

    @property
    def params(self) -> ParamList:
        return self.__params

    @property
    def body(self) -> ModBody:
        return self.__body

    def __str__(self) -> str:
        return "fmod"\
               + " " + self.name\
               + str(self.params)\
               + " " + "is"\
               + "\n" + indent(str(self.body))\
               + "\n" + "endfm"


class AEModule(Module):
    def __init__(self, name: str, params: ParamList, body: ModBody) -> None:
        self.__name = name
        self.__params = params
        self.__body = body

    @property
    def name(self) -> str:
        return self.__name

    @property
    def params(self) -> ParamList:
        return self.__params

    @property
    def body(self) -> ModBody:
        return self.__body

    def __str__(self) -> str:
        return "aemod"\
               + " " + self.name\
               + str(self.params)\
               + " " + "is"\
               + "\n" + indent(str(self.body))\
               + "\n" + "endaem"


class EModule(Module):
    def __init__(self, name: str, params: ParamList, body: ModBody) -> None:
        self.__name = name
        self.__params = params
        self.__body = body

    @property
    def name(self) -> str:
        return self.__name

    @property
    def params(self) -> ParamList:
        return self.__params

    @property
    def body(self) -> ModBody:
        return self.__body

    def __str__(self) -> str:
        return "emod"\
               + " " + self.name\
               + str(self.params)\
               + " " + "is"\
               + "\n" + indent(str(self.body))\
               + "\n" + "endem"


class ThBody:
    def __init__(self, *instrs: Instruction) -> None:
        self.__instrs = list(instrs)

    @property
    def instrs(self) -> List[Instruction]:
        return self.__instrs

    def __str__(self) -> str:
        return "\n".join([str(x) for x in self.instrs])


class Theory(MaudeElem):
    def __init__(self, name: str, body: ThBody) -> None:
        self.__name = name
        self.__body = body

    @property
    def name(self) -> str:
        return self.__name

    @property
    def body(self) -> ThBody:
        return self.__body

    def __str__(self) -> str:
        return "th"\
               + " " + self.name\
               + " " + "is"\
               + "\n" + indent(str(self.body))\
               + "\n" + "endth"


class Mapping:
    @property
    def lhs(self) -> str:
        raise NotImplementedError("This should be implemented in subclasses.")

    @property
    def rhs(self) -> str:
        raise NotImplementedError("This should be implemented in subclasses.")


class ViewBody:
    def __init__(self, *maps: Mapping) -> None:
        self.__maps = list(maps)

    @property
    def maps(self) -> List[Mapping]:
        return self.__maps

    def __str__(self) -> str:
        return "\n".join([str(x) for x in self.maps])


class View(MaudeElem):
    def __init__(self, name: str, mod_from: str, mod_to: str, body: ViewBody) -> None:
        self.__name = name
        self.__mod_from = mod_from.strip()
        self.__mod_to = mod_to.strip()
        self.__body = body

    @property
    def name(self) -> str:
        return self.__name

    @property
    def mod_from(self) -> str:
        return self.__mod_from

    @property
    def mod_to(self) -> str:
        return self.__mod_to

    @property
    def body(self) -> ViewBody:
        return self.__body

    def __str__(self) -> str:
        return "view"\
               + " " + self.name\
               + " " + "from"\
               + " " + self.mod_from\
               + " " + "to" \
               + " " + self.mod_to\
               + " " + "is"\
               + "\n" + indent(str(self.body))\
               + "\n" + "endv"


class Protecting(Instruction):
    def __init__(self, mod_expr: ModExpr) -> None:
        self.__mod_expr = mod_expr

    @property
    def mod_expr(self) -> ModExpr:
        return self.__mod_expr

    def __str__(self) -> str:
        return "pr"\
               + " " + str(self.mod_expr)\
               + " " + "."


class Extending(Instruction):
    def __init__(self, mod_expr: ModExpr) -> None:
        self.__mod_expr = mod_expr

    @property
    def mod_expr(self) -> ModExpr:
        return self.__mod_expr

    def __str__(self) -> str:
        return "ex" \
               + " " + str(self.mod_expr) \
               + " " + "."


class Including(Instruction):
    def __init__(self, mod_expr: ModExpr) -> None:
        self.__mod_expr = mod_expr

    @property
    def mod_expr(self) -> ModExpr:
        return self.__mod_expr

    def __str__(self) -> str:
        return "inc" \
               + " " + str(self.mod_expr) \
               + " " + "."


class Attr:
    pass


class AttrList:
    def __init__(self, *attrs: Attr) -> None:
        self.__attrs = list(attrs)

    @property
    def attrs(self) -> List[Attr]:
        return self.__attrs

    def __str__(self) -> str:
        if len(self.attrs) > 0:
            return "["\
                   + " ".join([str(x) for x in self.attrs])\
                   + "]"
        else:
            return ""


class Rule(Instruction):
    def __init__(self, lhs: Term, rhs: Term, attrs: AttrList = AttrList())\
            -> None:
        self.__lhs = lhs
        self.__rhs = rhs
        self.__attrs = attrs

    @property
    def lhs(self) -> Term:
        return self.__lhs

    @property
    def rhs(self) -> Term:
        return self.__rhs

    @property
    def attrs(self) -> AttrList:
        return self.__attrs

    def __str__(self) -> str:
        return "rl"\
               + " " + str(self.lhs)\
               + " " + "=>"\
               + " " + str(self.rhs)\
               + " " + str(self.attrs)\
               + " " + "."


class AtomicCondition:
    pass


class Condition:
    def __init__(self, *conds: AtomicCondition) -> None:
        self.__conds = list(conds)

    @property
    def conds(self) -> List[AtomicCondition]:
        return self.__conds

    def __str__(self) -> str:
        return r" /\ ".join([str(c) for c in self.conds])


class CondRule(Instruction):
    def __init__(self, lhs: Term, rhs: Term, cond: Condition, attrs: AttrList = AttrList()) -> None:
        self.__lhs = lhs
        self.__rhs = rhs
        self.__cond = cond
        self.__attrs = attrs

    @property
    def lhs(self) -> Term:
        return self.__lhs

    @property
    def rhs(self) -> Term:
        return self.__rhs

    @property
    def cond(self) -> Condition:
        return self.__cond

    @property
    def attrs(self) -> AttrList:
        return self.__attrs

    def __str__(self) -> str:
        return "crl"\
               + " " + str(self.lhs)\
               + " " + "=>"\
               + " " + str(self.rhs)\
               + " " + "if"\
               + " " + str(self.cond)\
               + " " + str(self.attrs)\
               + " " + "."


class EgRule(Instruction):
    def __init__(self, lhs: Term, mid: Term, rhs: Term, attrs: AttrList = AttrList()) -> None:
        self.__lhs = lhs
        self.__mid = mid
        self.__rhs = rhs
        self.__attrs = attrs

    @property
    def lhs(self) -> Term:
        return self.__lhs

    @property
    def mid(self) -> Term:
        return self.__mid

    @property
    def rhs(self) -> Term:
        return self.__rhs

    @property
    def attrs(self) -> AttrList:
        return self.__attrs

    def __str__(self) -> str:
        return "rl"\
               + " " + str(self.lhs)\
               + " " + "=["\
               + " " + str(self.mid)\
               + " " + "]=>"\
               + " " + str(self.rhs)\
               + " " + str(self.attrs)\
               + " " + "."


class CondEgRule(Instruction):
    def __init__(self, lhs: Term, mid: Term, rhs: Term, cond: Condition, attrs: AttrList = AttrList()) -> None:
        self.__lhs = lhs
        self.__mid = mid
        self.__rhs = rhs
        self.__cond = cond
        self.__attrs = attrs

    @property
    def lhs(self) -> Term:
        return self.__lhs

    @property
    def mid(self) -> Term:
        return self.__mid

    @property
    def rhs(self) -> Term:
        return self.__rhs

    @property
    def cond(self) -> "Condition":
        return self.__cond

    @property
    def attrs(self) -> AttrList:
        return self.__attrs

    def __str__(self) -> str:
        return "crl"\
               + " " + str(self.lhs)\
               + " " + "=["\
               + " " + str(self.mid)\
               + " " + "]=>"\
               + " " + str(self.rhs)\
               + " " + "if"\
               + " " + str(self.cond)\
               + " " + str(self.attrs)\
               + " " + "."


class IdList:
    def __init__(self, *ids: str) -> None:
        self.__ids = list([i.strip() for i in ids])

    @property
    def ids(self) -> List[str]:
        return self.__ids

    def __str__(self) -> str:
        return " ".join([str(x) for x in self.ids])


class SortDecl(Instruction):
    def __init__(self, sorts: IdList) -> None:
        self.__sorts = sorts

    @property
    def sorts(self) -> IdList:
        return self.__sorts

    def __str__(self) -> str:
        return "sort"\
               + " " + str(self.sorts)\
               + " " + "."


class SubsortBody:
    def __init__(self, *sort_lists: IdList) -> None:
        self.__sort_lists = list(sort_lists)

    @property
    def sort_lists(self) -> List[IdList]:
        return self.__sort_lists

    def __str__(self) -> str:
        return (" " + "<" + " ").join([str(x) for x in self.sort_lists])


class SubsortDecl(Instruction):
    def __init__(self, body: SubsortBody) -> None:
        self.__body = body

    @property
    def body(self) -> SubsortBody:
        return self.__body

    def __str__(self) -> str:
        return "subsort"\
               + " " + str(self.body)\
               + " " + "."


class OpDecl(Instruction):
    def __init__(self, ops: IdList, from_sorts: IdList, arrow: str, to_sort: str, attrs: AttrList = AttrList()) -> None:
        self.__ops = ops
        self.__from_sorts = from_sorts
        self.__arrow = arrow.strip()
        self.__to_sort = to_sort.strip()
        self.__attrs = attrs

    @property
    def ops(self) -> IdList:
        return self.__ops

    @property
    def from_sorts(self) -> IdList:
        return self.__from_sorts

    @property
    def arrow(self) -> str:
        return self.__arrow

    @property
    def to_sort(self) -> str:
        return self.__to_sort

    @property
    def attrs(self) -> IdList:
        return self.__attrs

    def __str__(self) -> str:
        return "ops"\
               + " " + str(self.ops)\
               + " " + ":"\
               + " " + str(self.from_sorts)\
               + " " + self.arrow\
               + " " + self.to_sort\
               + " " + str(self.attrs)\
               + " " + "."


class Eq(Instruction):
    def __init__(self, lhs: Term, rhs: Term, attrs: AttrList = AttrList()) -> None:
        self.__lhs = lhs
        self.__rhs = rhs
        self.__attrs = attrs

    @property
    def lhs(self) -> Term:
        return self.__lhs

    @property
    def rhs(self) -> Term:
        return self.__rhs

    @property
    def attrs(self) -> AttrList:
        return self.__attrs

    def __str__(self) -> str:
        return "eq"\
               + " " + str(self.lhs)\
               + " " + "="\
               + " " + str(self.rhs)\
               + " " + str(self.attrs)\
               + " " + "."


class AttrComm(Attr):
    def __str__(self) -> str:
        return 'comm'


class AttrAssoc(Attr):
    def __str__(self) -> str:
        return 'assoc'


class AttrOwise(Attr):
    def __str__(self) -> str:
        return 'owise'


class AttrId(Attr):
    def __init__(self, term: str) -> None:
        self.__term = term

    @property
    def term(self) -> str:
        return self.__term

    def __str__(self) -> str:
        return "id:"\
               + " " + self.term


class AttrCtor(Attr):
    def __str__(self) -> str:
        return 'ctor'


class CondEq(Instruction):
    def __init__(self, lhs: Term, rhs: Term, cond: Condition, attrs: AttrList = AttrList()) -> None:
        self.__lhs = lhs
        self.__rhs = rhs
        self.__cond = cond
        self.__attrs = attrs

    @property
    def lhs(self) -> Term:
        return self.__lhs

    @property
    def rhs(self) -> Term:
        return self.__rhs

    @property
    def cond(self) -> Condition:
        return self.__cond

    @property
    def attrs(self) -> AttrList:
        return self.__attrs

    def __str__(self) -> str:
        return "ceq"\
               + " " + str(self.lhs)\
               + " " + "="\
               + " " + str(self.rhs)\
               + " " + "if"\
               + " " + str(self.cond)\
               + " " + str(self.attrs)\
               + " " + "."


class VarDecl(Instruction):
    def __init__(self, ids: IdList, sort: str) -> None:
        self.__ids = ids
        self.__sort = sort.strip()

    @property
    def ids(self) -> IdList:
        return self.__ids

    @property
    def sort(self) -> str:
        return self.__sort

    def __str__(self) -> str:
        return "var"\
               + " " + str(self.ids)\
               + " " + ":"\
               + " " + str(self.sort)\
               + " " + "."


class PptDecl(Instruction):
    def __init__(self, ppts: IdList, from_sorts: IdList, arrow: str, to_sort: str) -> None:
        self.__ppts = ppts
        self.__from_sorts = from_sorts
        self.__arrow = arrow.strip()
        self.__to_sort = to_sort.strip()

    @property
    def ppts(self) -> IdList:
        return self.__ppts

    @property
    def from_sorts(self) -> IdList:
        return self.__from_sorts

    @property
    def arrow(self) -> str:
        return self.__arrow

    @property
    def to_sort(self) -> str:
        return self.__to_sort

    def __str__(self) -> str:
        return "ppt"\
               + " " + str(self.ppts)\
               + " " + ":"\
               + " " + str(self.from_sorts)\
               + " " + self.arrow\
               + " " + str(self.to_sort)\
               + " " + "."


class Crit:
    @property
    def lhs(self) -> str:
        raise NotImplementedError("This should be implemented in subclasses.")

    @property
    def rhs(self) -> str:
        raise NotImplementedError("This should be implemented in subclasses.")


class CritList:
    def __init__(self, *crits: Crit) -> None:
        self.__crits = crits

    @property
    def crits(self) -> List[Crit]:
        return list(self.__crits)

    def __str__(self) -> str:
        return (" " + "/\\" + " ").join([str(x) for x in self.crits])


class ModList:
    def __init__(self, *ids: str) -> None:
        self.__ids = list([i.strip() for i in ids])

    @property
    def ids(self) -> List[str]:
        return self.__ids

    def __str__(self) -> str:
        return (" " + "||" + " ").join([str(x) for x in self.ids])


class SyncOn(Instruction):
    def __init__(self, mods: ModList, crits: CritList) -> None:
        self.__mods = mods
        self.__crits = crits

    @property
    def mods(self) -> ModList:
        return self.__mods

    @property
    def crits(self) -> CritList:
        return self.__crits

    def __str__(self) -> str:
        return "sync"\
               + " " + str(self.mods)\
               + " " + "on"\
               + " " + str(self.crits)\
               + " " + "."


class InhStat(Instruction):
    def __init__(self, lhs: str, rhs: str):
        self.__lhs = lhs.strip()
        self.__rhs = rhs.strip()

    @property
    def lhs(self) -> str:
        return self.__lhs

    @property
    def rhs(self) -> str:
        return self.__rhs

    def __str__(self) -> str:
        return "inh"\
               + " " + str(self.lhs)\
               + " " + "="\
               + " " + str(self.rhs)\
               + " " + "."


class Formula:
    pass


class AGStat(Instruction):
    def __init__(self, asm: Formula, grt: Formula) -> None:
        self.__asm = asm
        self.__grt = grt

    @property
    def asm(self) -> Formula:
        return self.__asm

    @property
    def grt(self) -> Formula:
        return self.__grt

    def __str__(self) -> str:
        return "ag"\
               + " " + str(self.asm)\
               + " " + "|>"\
               + " " + str(self.grt)\
               + " " + "."


class Mb(Instruction):
    def __init__(self, term: Term, sort: str) -> None:
        self.__term = term
        self.__sort = sort.strip()

    @property
    def term(self) -> Term:
        return self.__term

    @property
    def sort(self) -> str:
        return self.__sort

    def __str__(self) -> str:
        return "mb"\
                + " " + str(self.term)\
                + " " + ":"\
                + " " + self.sort\
                + " " + "."


class CondMb(Instruction):
    def __init__(self, term: Term, sort: str, cond: Condition) -> None:
        self.__term = term
        self.__sort = sort.strip()
        self.__cond = cond

    @property
    def term(self) -> Term:
        return self.__term

    @property
    def sort(self) -> str:
        return self.__sort

    @property
    def cond(self) -> Condition:
        return self.__cond

    def __str__(self) -> str:
        return "cmb"\
                + " " + str(self.term)\
                + " " + ":"\
                + " " + self.sort\
                + " " + "if"\
                + " " + str(self.cond)\
                + " " + "."


class ModExprId(ModExpr):
    def __init__(self, name: str) -> None:
        self.__name = name.strip()

    @property
    def name(self) -> str:
        return self.__name

    def __str__(self) -> str:
        return self.name


class ArgList:
    def __init__(self, *ids: str) -> None:
        self.__ids = list([x.strip() for x in ids])

    @property
    def ids(self) -> List[str]:
        return self.__ids

    def __str__(self) -> str:
        return ("," + " ").join([str(x) for x in self.ids])


class ModExprWArgs(ModExpr):
    def __init__(self, name: str, args: ArgList) -> None:
        self.__name = name.strip()
        self.__args = args

    @property
    def name(self) -> str:
        return self.__name

    @property
    def args(self) -> ArgList:
        return self.__args

    def __str__(self) -> str:
        return self.name\
               + "{"\
               + ", ".join(self.args.ids)\
               + "}"


class Renaming:
    pass


class RenamingList:
    def __init__(self, *rens: Renaming) -> None:
        self.__rens = list(rens)

    @property
    def rens(self) -> List[Renaming]:
        return self.__rens

    def __str__(self) -> str:
        return ", ".join([str(x) for x in self.rens])


class ModExprWRen(ModExpr):
    def __init__(self, name: str, rens: RenamingList) -> None:
        self.__name = name.strip()
        self.__rens = rens

    @property
    def name(self) -> str:
        return self.__name

    @property
    def rens(self) -> RenamingList:
        return self.__rens

    def __str__(self) -> str:
        return self.name\
               + " " + "*"\
               + " " + "("\
               + str(self.rens)\
               + ")"


class SortRenaming(Renaming):
    def __init__(self, lhs: str, rhs: str) -> None:
        self.__lhs = lhs.strip()
        self.__rhs = rhs.strip()

    @property
    def lhs(self) -> str:
        return self.__lhs

    @property
    def rhs(self) -> str:
        return self.__rhs

    def __str__(self) -> str:
        return "sort"\
               + " " + self.lhs\
               + " " + "to"\
               + " " + self.rhs


class OpRenaming(Renaming):
    def __init__(self, lhs: str, rhs: str) -> None:
        self.__lhs = lhs.strip()
        self.__rhs = rhs.strip()

    @property
    def lhs(self) -> str:
        return self.__lhs

    @property
    def rhs(self) -> str:
        return self.__rhs

    def __str__(self) -> str:
        return "op"\
               + " " + self.lhs\
               + " " + "to"\
               + " " + self.rhs


class EqualCondition(AtomicCondition):
    def __init__(self, lhs: Term, rhs: Term) -> None:
        self.__lhs = lhs
        self.__rhs = rhs

    @property
    def lhs(self) -> Term:
        return self.__lhs

    @property
    def rhs(self) -> Term:
        return self.__rhs

    def __str__(self) -> str:
        return str(self.lhs)\
               + " " + "="\
               + " " + str(self.rhs)


class AssignCondition(AtomicCondition):
    def __init__(self, lhs: Term, rhs: Term) -> None:
        self.__lhs = lhs
        self.__rhs = rhs

    @property
    def lhs(self) -> Term:
        return self.__lhs

    @property
    def rhs(self) -> Term:
        return self.__rhs

    def __str__(self) -> str:
        return str(self.lhs)\
               + " " + ":="\
               + " " + str(self.rhs)


class MemberCondition(AtomicCondition):
    def __init__(self, term: Term, sort: str) -> None:
        self.__term = term
        self.__sort = sort

    @property
    def term(self) -> Term:
        return self.__term

    @property
    def sort(self) -> str:
        return self.__sort

    def __str__(self) -> str:
        return str(self.term)\
               + " " + ":"\
               + " " + self.sort


class RewCondition(AtomicCondition):
    def __init__(self, lhs: Term, rhs: Term) -> None:
        self.__lhs = lhs
        self.__rhs = rhs

    @property
    def lhs(self) -> Term:
        return self.__lhs

    @property
    def rhs(self) -> Term:
        return self.__rhs

    def __str__(self) -> str:
        return str(self.lhs)\
               + " " + "=>"\
               + " " + str(self.rhs)


class BoolCondition(AtomicCondition):
    def __init__(self, text: Term) -> None:
        self.__text = text

    @property
    def text(self) -> Term:
        return self.__text

    def __str__(self) -> str:
        return str(self.text)


class OpMapping(Mapping):
    def __init__(self, lhs: str, rhs: str) -> None:
        self.__lhs = lhs.strip()
        self.__rhs = rhs.strip()

    @property
    def lhs(self) -> str:
        return self.__lhs

    @property
    def rhs(self) -> str:
        return self.__rhs

    def __str__(self) -> str:
        return "op"\
               + " " + self.lhs\
               + " " + "to"\
               + " " + self.rhs\
               + " " + "."


class SortMapping(Mapping):
    def __init__(self, lhs: str, rhs: str) -> None:
        self.__lhs = lhs.strip()
        self.__rhs = rhs.strip()

    @property
    def lhs(self) -> str:
        return self.__lhs

    @property
    def rhs(self) -> str:
        return self.__rhs

    def __str__(self) -> str:
        return "sort"\
               + " " + self.lhs\
               + " " + "to"\
               + " " + self.rhs\
               + " " + "."


class PptMapping(Mapping):
    def __init__(self, lhs: str, rhs: str) -> None:
        self.__lhs = lhs.strip()
        self.__rhs = rhs.strip()

    @property
    def lhs(self) -> str:
        return self.__lhs

    @property
    def rhs(self) -> str:
        return self.__rhs

    def __str__(self) -> str:
        return "ppt"\
               + " " + self.lhs\
               + " " + "to"\
               + " " + self.rhs\
               + " " + "."


class EqCrit(Crit):
    def __init__(self, lhs: str, rhs: str) -> None:
        self.__lhs = lhs.strip()
        self.__rhs = rhs.strip()

    @property
    def lhs(self) -> str:
        return self.__lhs

    @property
    def rhs(self) -> str:
        return self.__rhs

    def __str__(self) -> str:
        return self.lhs\
               + " " + "="\
               + " " + self.rhs


class AssignCrit(Crit):
    def __init__(self, lhs: str, rhs: str) -> None:
        self.__lhs = lhs.strip()
        self.__rhs = rhs.strip()

    @property
    def lhs(self) -> str:
        return self.__lhs

    @property
    def rhs(self) -> str:
        return self.__rhs

    def __str__(self) -> str:
        return self.lhs\
               + " " + "="\
               + " " + self.rhs


class TrueFormula(Formula):
    def __init__(self) -> None:
        pass

    def __str__(self):
        return "True"


class FalseFormula(Formula):
    def __init__(self) -> None:
        pass

    def __str__(self):
        return "False"


class AtomicFormula(Formula):
    def __init__(self, text: str) -> None:
        self.__text = text.strip()

    @property
    def text(self) -> str:
        return self.__text

    def __str__(self):
        return self.text


class ParenFormula(Formula):
    def __init__(self, form: Formula) -> None:
        self.__form = form

    @property
    def form(self) -> Formula:
        return self.__form

    def __str__(self) -> str:
        return "("\
               + str(self.form)\
               + ")"


class NegFormula(Formula):
    def __init__(self, form: Formula) -> None:
        self.__form = form

    @property
    def form(self) -> Formula:
        return self.__form

    def __str__(self) -> str:
        return "~"\
               + " " + str(self.form)


class ConjFormula(Formula):
    def __init__(self, lhs: Formula, rhs: Formula) -> None:
        self.__lhs = lhs
        self.__rhs = rhs

    @property
    def lhs(self) -> Formula:
        return self.__lhs

    @property
    def rhs(self) -> Formula:
        return self.__rhs

    def __str__(self) -> str:
        return str(self.lhs)\
               + " " + "/\\"\
               + " " + str(self.rhs)


class DisjFormula(Formula):
    def __init__(self, lhs: Formula, rhs: Formula) -> None:
        self.__lhs = lhs
        self.__rhs = rhs

    @property
    def lhs(self) -> Formula:
        return self.__lhs

    @property
    def rhs(self) -> Formula:
        return self.__rhs

    def __str__(self) -> str:
        return str(self.lhs)\
               + " " + "\\/"\
               + " " + str(self.rhs)


class ImplFormula(Formula):
    def __init__(self, lhs: Formula, rhs: Formula) -> None:
        self.__lhs = lhs
        self.__rhs = rhs

    @property
    def lhs(self) -> Formula:
        return self.__lhs

    @property
    def rhs(self) -> Formula:
        return self.__rhs

    def __str__(self) -> str:
        return str(self.lhs)\
               + " " + "->"\
               + " " + str(self.rhs)


class EquivFormula(Formula):
    def __init__(self, lhs: Formula, rhs: Formula) -> None:
        self.__lhs = lhs
        self.__rhs = rhs

    @property
    def lhs(self) -> Formula:
        return self.__lhs

    @property
    def rhs(self) -> Formula:
        return self.__rhs

    def __str__(self) -> str:
        return str(self.lhs)\
               + " " + "<->"\
               + " " + str(self.rhs)


class ReleaseFormula(Formula):
    def __init__(self, lhs: Formula, rhs: Formula) -> None:
        self.__lhs = lhs
        self.__rhs = rhs

    @property
    def lhs(self) -> Formula:
        return self.__lhs

    @property
    def rhs(self) -> Formula:
        return self.__rhs

    def __str__(self) -> str:
        return str(self.lhs)\
               + " " + "R"\
               + " " + str(self.rhs)


class UntilFormula(Formula):
    def __init__(self, lhs: Formula, rhs: Formula) -> None:
        self.__lhs = lhs
        self.__rhs = rhs

    @property
    def lhs(self) -> Formula:
        return self.__lhs

    @property
    def rhs(self) -> Formula:
        return self.__rhs

    def __str__(self) -> str:
        return str(self.lhs)\
               + " " + "U"\
               + " " + str(self.rhs)


class WeakUntilFormula(Formula):
    def __init__(self, lhs: Formula, rhs: Formula) -> None:
        self.__lhs = lhs
        self.__rhs = rhs

    @property
    def lhs(self) -> Formula:
        return self.__lhs

    @property
    def rhs(self) -> Formula:
        return self.__rhs

    def __str__(self) -> str:
        return str(self.lhs)\
               + " " + "W"\
               + " " + str(self.rhs)


class BoxFormula(Formula):
    def __init__(self, form: Formula) -> None:
        self.__form = form

    @property
    def form(self) -> Formula:
        return self.__form

    def __str__(self) -> str:
        return "[]"\
               + " " + str(self.form)


class DiamFormula(Formula):
    def __init__(self, form: Formula) -> None:
        self.__form = form

    @property
    def form(self) -> Formula:
        return self.__form

    def __str__(self) -> str:
        return "<>"\
               + " " + str(self.form)
