#!/usr/bin/env python3.9


import itertools
from typing import Any, Union, Tuple
from functools import singledispatch
import logging
from lark import Token, Tree
from objmodel import *
from qualifier import Qualifier
from inliner import Inliner
from tools import buildName, dollarToName, buildPptSortName,\
    split_dollarized_id, remove_dupls, symbolToClassName, freshVarName,\
    makeCompStg, compStgName
from atomizer import Atomizer
from condprocessor import CondProcessor
from stdizer import Stdizer
from bodyprocessor import BodyProcessor
from topmethods import get_mod_by_name


def sync_on_as_std_partit_trans(top: MaudeTop, so: SyncOn) -> List[Instruction]:
    if so is None:
        return []
    res = []
    modNames = so.mods.ids
    mods = [get_mod_by_name(top, mn) for mn in modNames]
    fmods = [sys_flat_mod(top, m) for m in mods]
    sfmods = [make_mod_std_partit_trans(top, fm, False) for fm in fmods]
    fsfmods = [inc_flat(top, sfm) for sfm in sfmods]
    qfsfmods = [Qualifier.qualify_module(fsfm, fsfm.name, top)
                for fsfm in fsfmods]
    bps = [BodyProcessor(m.body) for m in qfsfmods]
    res += simplifyImports([i for bp in bps for i in bp.all_imports])
    res += simplifySortDecls([sd for bp in bps for sd in bp.sort_decls])
    res += [ssd for bp in bps for ssd in bp.subsort_decls]
    # TODO: Subsorts may be repeated: simplify.
    res += [od for bp in bps for od in bp.op_decls]
    res += [vd for bp in bps for vd in bp.var_decls]
    res += [mb for bp in bps for mb in bp.mbs]
    res += [cmb for bp in bps for cmb in bp.cond_mbs]
    res += [e for bp in bps for e in bp.eqs]
    res += [e for bp in bps for e in bp.cond_eqs]
    res += [cr for bp, n in zip(bps, modNames)
            for cr in controlled_rules(bp, n)]
    res += decl_composed_stages(so.mods.ids)
    res += declCompInit(so.mods.ids)
    res += top_rule_for_split(so)
    return res


def sync_on_as_std_comp_trans(top: MaudeTop, so: SyncOn) -> List[Instruction]:
    if so is None:
        return []
    res = []
    mods = [get_mod_by_name(top, mn) for mn in so.mods.ids]
    fmods = [sys_flat_mod(top, m) for m in mods]
    sfmods = [make_mod_std_comp_trans(top, fm, False) for fm in fmods]
    fsfmods = [inc_flat(top, sfm) for sfm in sfmods]
    qfsfmods = [Qualifier.qualify_module(fsfm, fsfm.name, top)
                for fsfm in fsfmods]
    bps = [BodyProcessor(m.body) for m in qfsfmods]
    res += simplifyImports([i for bp in bps for i in bp.all_imports])
    res += simplifySortDecls([sd for bp in bps for sd in bp.sort_decls])
    res += [ssd for bp in bps for ssd in bp.subsort_decls]
    # TODO: Subsorts may be repeated: simplify.
    res += [od for bp in bps for od in bp.op_decls]
    res += [vd for bp in bps for vd in bp.var_decls]
    res += [mb for bp in bps for mb in bp.mbs]
    res += [cmb for bp in bps for cmb in bp.cond_mbs]
    res += [e for bp in bps for e in bp.eqs]
    res += [e for bp in bps for e in bp.cond_eqs]
    cp = CondProcessor(top, FModule("<TEMP>", ParamList(), ModBody(*res)))
    res += decl_composed_stages(so.mods.ids)
    res += declCompInit(so.mods.ids)
    res += comp_rules_from_sync_on(
        so.mods.ids,
        [[rule_to_cond(r) for r in bp.all_non_eg_rules] for bp in bps],
        so.crits, cp)
    return res


def make_emod_std_for_verif(top: MaudeTop, em: EModule) -> SModule:
    bp = BodyProcessor(em.body)
    if bp.sync_on is None:
        return make_mod_std_comp_trans(top, em)
    if not do_all_comps_have_ag(top, bp.sync_on):
        return make_mod_std_comp_trans(top, em)
    instrs = []
    instrs += [inh_to_simple_eq(inh) for inh in bp.inh_stats]
    instrs += bp.all_imports
    instrs += bp.sort_decls
    instrs += bp.subsort_decls
    instrs += bp.op_decls
    instrs += bp.var_decls
    instrs += bp.mbs
    instrs += bp.cond_mbs
    instrs += [Eq(e.lhs, Term(dollarToName(str(e.rhs))), e.attrs)
               for e in bp.eqs]
    instrs += [CondEq(e.lhs, Term(dollarToName(str(e.rhs))), e.cond, e.attrs)
               for e in bp.cond_eqs]  # TODO: ugly
    instrs += [fd for pd in bp.ppt_decls
               for fd in ppt_decl_to_form_decls(pd)]
    instrs += [Protecting(ModExprId("SAT-SOLVER"))]
    syncedMods = [get_mod_by_name(top, mn)
                  for mn in bp.sync_on.mods.ids]
    flatSyncedMods = [sys_flat_mod(top, m) for m in syncedMods]
    for fm in flatSyncedMods:  # all assumed to be in the same input file
        fb = fm.body
        fpb = BodyProcessor(fb)
        ii = fpb.all_imports
        ii.append(Protecting(ModExprId("SAT-SOLVER")))
        ii += fpb.sort_decls
        ii += fpb.subsort_decls
        ii += [fd for pd in fpb.ppt_decls
               for fd in ppt_decl_to_form_decls(pd)]
        ii += ags_to_op_decls_and_eqs(fpb.ag_stats)
        m = FModule(fm.name, ParamList(), ModBody(*ii))
        qm = Qualifier.qualify_module(m, fm.name, top)
        qmbp = BodyProcessor(qm.body)
        instrs += qmbp.op_decls
        instrs += qmbp.eqs
        # fb = fm.body.toQ(fm.name, self)
        # pptNames = [pn for pd in fb.ppt_decls for pn in pd.pptIdList.ids]
        # instrs += [OpDecl(IdList(pn), IdList(), "->", "Formula")
        #            for pn in pptNames]
        # instrs += fm.body.agsToQAgEqs(fm.name)
    instrs += crit_list_to_eqs(bp.sync_on.crits)
    for (nr, ag) in enumerate(bp.ag_stats):
        instrs += agstat_to_form_decl_and_eq(ag, nr)
    ags_conj = []
    for fm in flatSyncedMods:
        nr_ags = len(BodyProcessor(fm.body).ag_stats)
        ags_conj.append(r" /\ ".join(
            [buildName(fm.name, "ag" + str(nr)) for nr in range(nr_ags)]))
    compAgConj = r" /\ ".join(ags_conj)
    for nr in range(len(bp.ag_stats)):
        ded = buildName("ded" + str(nr))
        ag = buildName("ag" + str(nr))
        instrs += [OpDecl(IdList(ded), IdList(), "->", "Formula")]
        instrs += [Eq(Term(ded), Term("(" + compAgConj + ") -> (" + ag
                                      + ")"))]
        logging.info(f"red in {em.name} : tautCheck({ded}) .")
    return SModule(em.name, em.params, ModBody(*instrs))


def make_mod_std_partit_trans(top: MaudeTop, m: Module, withAgs: bool = True) \
        -> Union[FModule, SModule]:
    """ There is no point in translating ags if this module is not meant to
    be vefiried by itself, that is, if it is only being standardized because
    it appears in a sync on instruction, and is going to be inserted into
    the standard version of the composed one. Also, the result of
    translating ags produced some (minor) problems, that are simply avoided
    forgetting about them.
    """
    bp = BodyProcessor(m.body)
    instrs = []
    instrs += ppt_decls_in_body_as_std(bp)
    if withAgs:
        instrs += ags_as_std(bp)
    instrs += sync_on_as_std_partit_trans(top, bp.sync_on)
    instrs += [inh_to_eval_eq(inh) for inh in bp.inh_stats]
    instrs += bp.all_imports
    instrs += bp.sort_decls
    instrs += bp.subsort_decls
    instrs += bp.op_decls
    instrs += bp.var_decls
    instrs += bp.mbs
    instrs += bp.cond_mbs
    instrs += [Eq(e.lhs, Term(dollarToName(str(e.rhs))), e.attrs)
               for e in bp.eqs]
    instrs += [CondEq(e.lhs, Term(dollarToName(str(e.rhs))), e.cond,
                      e.attrs)
               for e in bp.cond_eqs]  # TODO: ugly
    instrs += bp.all_non_eg_rules
    instrs += [r for er in bp.all_eg_rules
               for r in split_rule(er)]
    instrs += bp.comments
    instrs = simplifyImports(instrs)
    instrs = simplifySortDecls(instrs)
    if isinstance(m, FModule):
        return FModule(m.name, m.params, ModBody(*instrs))
    else:
        return SModule(m.name, m.params, ModBody(*instrs))


def make_mod_std_comp_trans(top: MaudeTop, m: Module, withAgs: bool = True) \
        -> Union[FModule, SModule]:
    bp = BodyProcessor(m.body)
    instrs = []
    instrs += ppt_decls_in_body_as_std(bp)
    if withAgs:
        instrs += ags_as_std(bp)
    instrs += sync_on_as_std_comp_trans(top, bp.sync_on)
    instrs += [inh_to_eval_eq(inh) for inh in bp.inh_stats]
    instrs += bp.all_imports
    instrs += bp.sort_decls
    instrs += bp.subsort_decls
    instrs += bp.op_decls
    instrs += bp.var_decls
    instrs += bp.mbs
    instrs += bp.cond_mbs
    instrs += [Eq(e.lhs, Term(dollarToName(str(e.rhs))), e.attrs)
               for e in bp.eqs]
    instrs += [CondEq(e.lhs, Term(dollarToName(str(e.rhs))), e.cond,
                      e.attrs)
               for e in bp.cond_eqs]  # TODO: ugly
    instrs += bp.all_non_eg_rules
    instrs += [r for er in bp.all_eg_rules
               for r in split_rule(er)]
    instrs = simplifyImports(instrs)
    instrs = simplifySortDecls(instrs)
    if isinstance(m, FModule):
        return FModule(m.name, m.params, ModBody(*instrs))
    else:
        return SModule(m.name, m.params, ModBody(*instrs))


def make_std_partit_trans(top: MaudeTop, e: MaudeElem) -> MaudeElem:
    # TODO: It is odd that the two calls below have different shapes.
    if isinstance(e, SModule)\
            or isinstance(e, EModule) \
            or isinstance(e, AEModule):
        return make_mod_std_partit_trans(top, e)
    else:  # FModule, Theory, View
        return Stdizer().to_std(e)


def make_std_for_verif(top: MaudeTop, e: MaudeElem) -> MaudeElem:
    if isinstance(e, Module):
        if isinstance(e, EModule):
            return make_emod_std_for_verif(top, e)
        else:
            return make_mod_std_comp_trans(top, e)
    elif isinstance(e, Theory):
        return Stdizer().to_std(e)
    elif isinstance(e, View):
        return Stdizer().to_std(e)
    else:
        raise Exception("Wrong class.")


def make_std_comp_trans(top: MaudeTop, e: MaudeElem) -> MaudeElem:
    # TODO: It is odd that the two calls below have different shapes.
    if isinstance(e, SModule)\
            or isinstance(e, EModule) \
            or isinstance(e, AEModule):
        return make_mod_std_comp_trans(top, e)
    else:  # FModule, Theory, View
        return Stdizer().to_std(e)


def sys_flat_mod(top: MaudeTop, m: Module) -> Module:
    instrs = []
    for i in m.body.instrs:
        if isinstance(i, Protecting) \
                or isinstance(i, Extending) \
                or isinstance(i, Including):
            m2 = get_mod_by_name(top, i.mod_expr.name)
            if isinstance(m2, SModule) \
                    or isinstance(m2, AEModule) \
                    or isinstance(m2, EModule):
                instrs += Inliner.inline_vars_in_body(
                    sys_flat_mod(top, m2).body).instrs
            else:
                instrs.append(i)
        else:
            instrs.append(i)
    return m.__class__(m.name, m.params, ModBody(*instrs))


def inc_flat(top: MaudeTop, m: Module) -> Module:
    instrs = []
    for i in m.body.instrs:
        if isinstance(i, Including):
            m2 = get_mod_by_name(top, i.mod_expr.name)
            instrs += sys_flat_mod(top, m2).body.instrs
        else:
            instrs.append(i)
    return m.__class__(m.name, m.params, ModBody(*instrs))


def do_all_comps_have_ag(top, so: SyncOn) -> bool:
    """If some of these components is an EModule in turn,
    and involves new, more basic components, these new components are not
    considered here. """
    return all(
        [len(BodyProcessor(sys_flat_mod(
            top, get_mod_by_name(top, n)).body).ag_stats) > 0
         for n in so.mods.ids])


def crit_to_eq(c: Crit) -> Eq:
    lm, lp = split_dollarized_id(c.lhs)
    rm, rp = split_dollarized_id(c.rhs)
    return Eq(Term(buildName(lm, lp)),
              Term(buildName(rm, rp)))


def crit_list_to_eqs(cl: CritList) -> List[Eq]:
    return [crit_to_eq(c) for c in cl.crits]


def crit_list_to_condition(cl: CritList, stageTerms: List[str]) -> Condition:
    res = []
    for c in cl.crits:
        lm, lp = split_dollarized_id(c.lhs)
        rm, rp = split_dollarized_id(c.rhs)
        varL, varR = "", ""
        for v in stageTerms:
            if lm == v.split("|")[0][1:]:
                varL = v
            if rm == v.split("|")[0][1:]:
                varR = v
        res += [EqualCondition(Term(buildName(lm, lp) + " @ " + varL),
                               Term(buildName(rm, rp) + " @ " + varR))]
    return Condition(*res)


def top_rule_for_split(so: SyncOn) -> List[Instruction]:
    # This method is only here to unclutter the one calling it.
    res = []
    # declare stage vars for top rule
    varsL, varsR = [], []
    for mn in so.mods.ids:
        fvn = freshVarName()
        varsL.append(buildName(mn, fvn))
        res += [VarDecl(IdList(buildName(mn, fvn)), buildName(mn, "Stage"))]
        fvn = freshVarName()
        varsR.append(buildName(mn, fvn))
        res += [VarDecl(IdList(buildName(mn, fvn)), buildName(mn, "Stage"))]
    # compute condition
    conds = []
    for (vl, vr, mn) in zip(varsL, varsR, so.mods.ids):
        conds += [RewCondition(buildName(mn, "<ctrlStg>") + "(" + vl
                               + ", false)",
                               buildName(mn, "<ctrlStg>") + "(" + vr
                               + ", true)")]
    condAux = "(" + " or ".join([f"{vl} =/= {vr}"
                                 for (vl, vr) in zip(varsL, varsR)]) + ")"
    conds += [BoolCondition(Term(condAux))]
    conds += crit_list_to_condition(so.crits, varsR).conds
    # the rule
    res += [CondRule(Term(makeCompStg(varsL)),
                     Term(makeCompStg(varsR)),
                     Condition(*conds))]
    return res


def ag_atoms_to_op_eq(bp: BodyProcessor) -> Tuple[List[OpDecl], List[Eq]]:
    od = []
    e = []
    atomizer = Atomizer()
    atoms = [a for ag in bp.ag_stats
             for a in remove_dupls(atomizer.atoms(ag.asm)
                                   + atomizer.atoms(ag.grt))]
    for a in atoms:
        od += [OpDecl(IdList(a), IdList(), "->", "Prop")]
        v = buildName(freshVarName()) + ":Stage"
        e += [Eq(Term(v + " |= " + a), Term(a + " @ " + v))]
    return od, e


def ags_as_std(bp: BodyProcessor) -> List[Instruction]:
    if len(bp.ag_stats) > 0:
        newName = buildName("MC", "State")
        ex = [Extending(ModExprWRen("MODEL-CHECKER",
                                    RenamingList(SortRenaming("State",
                                                              newName))))]
        ssd = [SubsortDecl(SubsortBody(IdList("Stage"), IdList(newName)))]
    else:
        ex = []
        ssd = []
    od1, eq1 = ag_atoms_to_op_eq(bp)
    od2, eq2 = ags_to_op_eq_red(bp)
    res = ex
    res += ssd
    res += od1
    res += od2
    res += eq1
    res += eq2
    return res


def ags_to_op_eq_red(bp: BodyProcessor)\
        -> Tuple[List[OpDecl], List[Eq]]:
    od = []
    e = []
    ags = bp.ag_stats
    for (nr, ag) in enumerate(ags):
        agn = buildName("ag" + str(nr))
        od += [OpDecl(IdList(agn), IdList(), "->", "Formula")]
        e += [Eq(Term(agn), Term("(" + str(ag.asm)
                                 + ") -> (" + str(ag.grt)
                                 + ")"))]
        logging.info(f"red modelCheck(init, {agn}) .")
    return od, e


@singledispatch
def rule_to_ctrl(_, arg) -> Union[Rule, CondRule]:
    raise NotImplementedError(f"There is no toCtrl method for {type(arg)}.")


@rule_to_ctrl.register
def _(r: Rule, q: str) -> Rule:
    if str(r.lhs).find("ctrlStg") != -1:  # already controlled
        return r
    else:
        return Rule(Term(buildName(q, "<ctrlStg>") + "(" + str(r.lhs)
                         + ", false)"),
                    Term(buildName(q, "<ctrlStg>") + "(" + str(r.rhs)
                         + ", true)"),
                    r.attrs)


@rule_to_ctrl.register
def _(cr: CondRule, q: str) -> CondRule:
    if str(cr.lhs).find("ctrlStg") != -1:  # already controlled
        return cr
    else:
        return CondRule(Term(buildName(q, "<ctrlStg>") + "(" + str(cr.lhs)
                        + ", false)"),
                        Term(buildName(q, "<ctrlStg>") + "(" + str(cr.rhs)
                        + ", true)"),
                        cr.cond, cr.attrs)


def controlled_rules(bp: BodyProcessor, q: str) -> List[Instruction]:
    # This method is only here to unclutter the one calling it.
    res = []
    ctrlStageSortName = buildName(q, "CtrlStage")
    res += [SortDecl(IdList(ctrlStageSortName))]
    res += [OpDecl(IdList(buildName(q, "<ctrlStg>")),
                   IdList(buildName(q, "Stage"), "Bool"),
                   "->",
                   ctrlStageSortName)]
    res += [rule_to_ctrl(r, q) for r in bp.all_non_eg_rules]
    vn = freshVarName()
    res += [Rule(Term(buildName(q, "<ctrlStg>") + "(" + vn + ":"
                 + buildName(q, "Stage") + ", false)"),
                 Term(buildName(q, "<ctrlStg>") + "(" + vn + ":"
                 + buildName(q, "Stage") + ", true)"))]
    return res


def rule_to_cond(r: Union[Rule, CondRule]) -> CondRule:
    if isinstance(r, CondRule):
        return r
    else:
        return CondRule(Term(str(r.lhs)), Term(str(r.rhs)),
                        Condition(BoolCondition(Term('true'))),
                        r.attrs)


def reduce_conds(cr: CondRule, cp: CondProcessor) \
            -> Union[CondRule, Rule, None]:
    previousConds = []
    matchingConds = []
    equationalConds = []
    for e in cr.cond.conds:
        if isinstance(e, EqualCondition):
            rel = "="
            ll, rr = e.lhs, e.rhs
        elif isinstance(e, AssignCondition):
            rel = ":="
            ll, rr = e.lhs, e.rhs
        else:
            assert (isinstance(e, BoolCondition))
            rel = "="
            ll, rr = e.text, "true"
        if ll == rr:  # remove true = true conds (that I myself include)
            continue
        if " @ " not in str(ll) + str(rr):
            # only simplify conds from sync crit
            previousConds.append(e)
            continue
        redL = cp.reduce(str(ll))
        if " @ " in redL and rel == ":=":
            raise Exception("Non-reducible ppt expression in criterion.")
        redR = cp.reduce(str(rr))
        # if " @ " in redR:
        #     raise Exception("Non-reducible ppt expression in criterion.")
        if redL == redR:
            continue  # true condition
        if all([(v not in redL + redR)
                for vd in BodyProcessor(cp.mod.body).var_decls
                for v in vd.ids.ids]) and ":" not in redL + redR:
            # does not contain vbles (and redL!=redR)
            return None  # false condition
        if rel == "=":
            equationalConds.append(EqualCondition(Term(redL), Term(redR)))
        else:
            assert (rel == ":=")
            matchingConds.append(AssignCondition(Term(redL), Term(redR)))
        # TODO: I'm not sure this could be useful or wasteful:
        #  _, redEq = self.__maudeSession.reduce(res)
    allConds = []
    allConds += matchingConds
    allConds += previousConds
    allConds += equationalConds
    if len(allConds) == 0:
        return Rule(cr.lhs, cr.rhs, cr.attrs)
    else:
        return CondRule(cr.lhs, cr.rhs, Condition(*allConds),
                        cr.attrs)


def comp_rules_from_sync_on(modNames: List[str], rules: List[List[CondRule]],
                            criteria: CritList, cp: CondProcessor) \
        -> List[Instruction]:
    """
    This produces composed rules for composed stages from each system's
    rules. The conditions for each produced rule are all the conditions from
    each component rule, plus the ones resulting from syncing criteria.
    However, the conditions from syncing criteria are not complete: they
    lack the testing of whether each property is defined at the given stage
    or not (if it is not, the condition has to be considered trivially
    true). This cannot be done here, so it is delayed until all modules have
    been split and reductions of expressions can be performed (calling
    Maude). This is done in the class `condProcessor`.
    """
    res = []
    extRules = []
    # add to each set of rules a do-nothing rule
    flag = "dOn0tHiNg"
    for mn, rr in zip(modNames, rules):
        vn = buildName(mn, flag, freshVarName()) + ":"\
            + buildName(mn, "Stage")
        extRules += [rr + [CondRule(
            Term(vn), Term(vn), Condition(BoolCondition(Term("true"))))]]
    for rs in itertools.product(*extRules):
        lhsList, rhsList, condEltList = [], [], []
        for r in rs:
            assert (isinstance(r, CondRule))
            lhsList.append(str(r.lhs))
            rhsList.append(str(r.rhs))
            condEltList += r.cond.conds
        lhs = makeCompStg(lhsList)
        rhs = makeCompStg(rhsList)
        if lhs == rhs:  # avoid the case of all do-nothing rules
            continue
        cr = []
        for c in criteria.crits:
            # This needs quite an explanation.
            lm, lp = split_dollarized_id(c.lhs)
            rm, rp = split_dollarized_id(c.rhs)
            lStageTerm = rhsList[modNames.index(lm)]
            if flag in lStageTerm and isinstance(c, AssignCrit):
                lStageTerm = lhsList[modNames.index(rm)]
                rStageTerm = rhsList[modNames.index(rm)]
                cr.append(EqualCondition(Term(buildName(rm, rp)
                                         + " @ (" + lStageTerm + ")"),
                                         Term(buildName(rm, rp)
                                         + " @ (" + rStageTerm + ")")))
                continue
            lStageTerm = rhsList[modNames.index(lm)]
            rStageTerm = rhsList[modNames.index(rm)]
            if isinstance(c, EqCrit):
                cr.append(EqualCondition(Term(buildName(lm, lp)
                                         + " @ (" + lStageTerm + ")"),
                                         Term(buildName(rm, rp)
                                         + " @ (" + rStageTerm + ")")))
            else:
                assert isinstance(c, AssignCrit)
                cr.append(AssignCondition(Term(buildName(lm, lp)
                                          + " @ (" + lStageTerm + ")"),
                                          Term(buildName(rm, rp)
                                          + " @ (" + rStageTerm + ")")))
            lStageTerm = lhsList[modNames.index(lm)]
            rStageTerm = lhsList[modNames.index(rm)]
            if isinstance(c, EqCrit):
                cr.append(EqualCondition(Term(buildName(lm, lp)
                                         + " @ (" + lStageTerm + ")"),
                                         Term(buildName(rm, rp)
                                         + " @ (" + rStageTerm + ")")))
            else:
                assert isinstance(c, AssignCrit)
                cr.append(AssignCondition(Term(buildName(lm, lp)
                                          + " @ (" + lStageTerm + ")"),
                                          Term(buildName(rm, rp)
                                          + " @ (" + rStageTerm + ")")))
        conds = condEltList + cr
        r = reduce_conds(CondRule(Term(lhs), Term(rhs), Condition(*conds)), cp)
        if r is not None:
            res.append(r)
    return res


@singledispatch
def split_rule(arg):
    raise NotImplementedError(f"No split_rule method for {type(arg)}.")


@split_rule.register
def _(er: EgRule) -> List[Rule]:
    return [Rule(er.lhs, er.mid, er.attrs),
            Rule(er.mid, er.rhs, er.attrs)]


@split_rule.register
def _(cer: CondEgRule) -> List[CondRule]:
    return [CondRule(cer.lhs, cer.mid, cer.cond, cer.attrs),
            CondRule(cer.mid, cer.rhs, cer.cond, cer.attrs)]


def ppt_decl_to_form_decls(pd: PptDecl) -> List[OpDecl]:
    res = []
    for i in pd.ppts.ids:
        res += [OpDecl(IdList(i), pd.from_sorts, pd.arrow, "Formula")]
    return res


def ppt_decl_to_op_decls(pd: PptDecl) -> List[OpDecl]:
    res = []
    for i in pd.ppts.ids:
        res += [OpDecl(IdList(i), pd.from_sorts, pd.arrow,
                       buildPptSortName(pd.to_sort))]
    return res


def ppt_decls_in_body_as_std(bp: BodyProcessor) -> List[Instruction]:
    toSorts = remove_dupls([pd.to_sort for pd in bp.ppt_decls])
    pptSorts = [buildPptSortName(ts) for ts in toSorts]
    sortDecls = [SortDecl(IdList(ps)) for ps in pptSorts]
    opDecls = [OpDecl(IdList("_@_"), IdList(ps, "Stage"), "->", ts)
               for (ps, ts) in zip(pptSorts, toSorts)]
    opDecls += [od for pd in bp.ppt_decls
                for od in ppt_decl_to_op_decls(pd)]
    res = sortDecls
    res += opDecls
    return res


def inh_to_eval_eq(i: InhStat) -> Eq:
    vn = freshVarName() + ":Stage"
    m, p = split_dollarized_id(i.rhs)
    return Eq(Term(f'{i.lhs} @ {vn}'),
              Term(f'{buildName(m, p)} @ {m}({vn})'))


def inh_to_simple_eq(i: InhStat) -> Eq:
    m, p = split_dollarized_id(i.rhs)
    return Eq(Term(i.lhs), Term(buildName(m, p)))


def agstat_to_form_decl_and_eq(a: AGStat, nr: int) -> List[Instruction]:
    res = []
    res += [OpDecl(IdList(buildName("ag" + str(nr))), IdList(), "->",
                   "Formula")]
    res += [Eq(Term(buildName("ag" + str(nr))),
               Term("(" + str(a.asm) + ") -> (" + str(a.grt) + ")"))]
    return res


stageMod = SModule("STAGE", ParamList(),
                   ModBody(SortDecl(IdList("State", "Trans", "Stage")),
                           SubsortDecl(SubsortBody(IdList("State", "Trans"),
                                                   IdList("Stage"))),
                           OpDecl(IdList("init"), IdList(), "->", "Stage")
                           )
                   )


def astToObjModel(ast: Union[Token, Tree]) -> Any:
    if isinstance(ast, Token):
        return str(ast).strip()
    chldrn = [astToObjModel(c) for c in ast.children]
    classObj = globals()[symbolToClassName(ast.data)]
    return classObj(*chldrn)


def simplifyImports(ii: List[Instruction]) -> List[Instruction]:
    # This
    # removes duplicates and simplifies them. For example, `pr M` is removed if
    # `ex M` is present.
    # TODO: Make this a static method of the only class that uses it.
    res = []
    prs, exs, incs = [], [], []
    for i in ii:
        if isinstance(i, Protecting):
            prs.append(i)
        elif isinstance(i, Extending):
            exs.append(i)
        elif isinstance(i, Including):
            incs.append(i)
        else:
            res.append(i)
    resPrs, resExs, resIncs = [], [], []
    for p in remove_dupls(prs, lambda x: x.mod_expr.name):
        exsincs = exs
        exsincs += incs
        if p.mod_expr.name not in [x.mod_expr.name for x in exsincs]:
            resPrs.append(p)
    for e in remove_dupls(exs, lambda x: x.mod_expr.name):
        if e.mod_expr.name not in [x.mod_expr.name for x in incs]:
            resExs.append(e)
    resIncs = remove_dupls(incs, lambda x: x.mod_expr.name)
    # If there are repeated ids, maybe with different args or renamings,
    #  all this chooses one at random.
    res += resPrs
    res += resExs
    res += resIncs
    # Concatenating the three lists at once produces a Pycharm warning
    # because of typing. Not grave, but I prefer to avoid warnings.
    return res


def simplifySortDecls(ii: List[Instruction]) -> List[Instruction]:
    # TODO: Make this a static method of the only class that uses it.
    # Just remove duplicates
    allSortNames = []
    res = []
    for i in ii:
        if isinstance(i, SortDecl):
            allSortNames += i.sorts.ids
        else:
            res.append(i)
    nonDuplSortNames = remove_dupls(allSortNames)
    return res + [SortDecl(IdList(sn)) for sn in nonDuplSortNames]


def ags_to_op_decls_and_eqs(ags: List[AGStat]) -> List[Instruction]:
    res = []
    for nr, ag in enumerate(ags):
        n = 'ag' + str(nr)
        res += [OpDecl(IdList(n), IdList(), '->', 'Formula')]
        res += [Eq(Term(n), Term(f'({str(ag.asm)}) -> ({str(ag.grt)})'))]
    return res


def decl_composed_stages(modNames: List[str]) -> List[Instruction]:
    # This method is only here to unclutter the one calling it.
    res = [SortDecl(IdList("Stage"))]
    stageSortNames = [buildName(mn, "Stage") for mn in modNames]
    stageVarNames = [freshVarName() + ":" + ssn for ssn in stageSortNames]
    opName = compStgName()
    res += [OpDecl(IdList(opName), IdList(*stageSortNames), "->", "Stage")]
    # projection operators
    for (mn, sn, vn) in zip(modNames, stageSortNames, stageVarNames):
        res += [OpDecl(IdList(mn), IdList("Stage"), "->", sn)]
        res += [Eq(mn + "(" + opName + "(" + ", ".join(stageVarNames)
                   + "))", vn)]
    return res


def declCompInit(modNames: List[str]) -> List[Instruction]:
    # This method is only here to unclutter the one calling it.
    res = [OpDecl(IdList("init"), IdList(), "->", "Stage")]
    initEqRhs = makeCompStg([buildName(mn, "init") for mn in modNames])
    res += [Eq(Term("init"), Term(initEqRhs))]
    return res


def paramEmodImports(em: EModule)\
        -> List[Union[Protecting, Extending, Including]]:
    res = []
    for i in em.body.instrs:
        if (isinstance(i, Protecting) or isinstance(i, Extending)
            or isinstance(i, Including)) \
                and isinstance(i.mod_expr, ModExprWArgs):
            res.append(i)
    return res
