#!/usr/bin/env python3.9


from functools import singledispatchmethod
from typing import Dict
import maude
from objmodel import *
from tools import buildName
from bodyprocessor import BodyProcessor
from topmethods import get_context_str


class Qualifier:
    def __init__(self, body: ModBody, modName: str, top: MaudeTop) -> None:
        bp = BodyProcessor(body)
        signBody = ModBody(*(bp.all_imports + bp.all_decls))
        signName = buildName(modName, 'SIGN')
        signModStr = f'fmod {signName} is {str(signBody)} endfm'
        self.__qDict = dict()
        for n in bp.get_declared_names():
            if n == '_@_':
                self.__qDict['_@_'] = '_@_'
            else:
                self.__qDict[n] = buildName(modName, n)
        qSignBody = Qualifier.get_translated_signature(bp, self.__qDict)
        qSignName = buildName('Q', modName, 'SIGN')
        qSignModStr = f'fmod {qSignName} is {str(qSignBody)} endfm'
        contextStr = get_context_str(top, signBody.instrs, [])
        maude.init(advise=False)
        maude.load('model-checker.maude')
        maude.input(contextStr)
        maude.input(signModStr)
        maude.input(qSignModStr)
        self.__rrSignMod = maude.getModule(signName)
        self.__rrQSignMod = maude.getModule(qSignName)
        self.__q = modName

    @staticmethod
    def qualify_module(mod: Module, q: str, top: MaudeTop) -> Module:
        return mod.__class__(mod.name, mod.params,
                             Qualifier.__qualify_body(mod.body, q, top))

    @staticmethod
    def __qualify_body(body: ModBody, q: str, top: MaudeTop) -> ModBody:
        qr = Qualifier(body, q, top)
        qInstrs = []
        for i in body.instrs:
            qInstrs.append(qr.__qualify(i))
        return ModBody(*qInstrs)

    @staticmethod
    def get_translated_signature(bp: BodyProcessor, d: Dict[str, str])\
            -> ModBody:
        """This takes the signature part of `self` (that is,
        declarations of sorts, subsorts, vars, and operators, and also
        all imports) and produces a qualified equivalent by means of
        the received translation dictionary `qDict`. If `self` is not
        just a signature (it contains instructions), this only considers
        its signature part. The dictionary contains the correspondence
        of each qualifiable name to its qualified version. We use the
        `get` method for the dictionary, instead of the `[]` access, to
        account for the cases in which the id is not in the dictionary,
        in which case we leave the id as it was. The dictionary is
        expected to contain the correspondence of nonqualified with
        qualified names, but this function works with whatever values
        are in it. Note that this takes not a body, but a `BodyProcessor`.
        The only reason is that otherwise we would need to create such
        a `BodyProcessor` here in this method, and it the (unique)
        calling method it is already created.
        """
        instrs = []
        instrs += bp.all_imports
        # declare translated sorts
        sortNames = bp.get_declared_sorts()
        for sn in sortNames:
            qsn = d.get(sn, sn)
            instrs.append(SortDecl(IdList(qsn)))
        # declare subsorts for translated sorts
        for ss in bp.get_subsort_items():
            # sorts decl'd in imported modules are not in the dict
            qss0 = d.get(ss[0], ss[0])
            qss1 = d.get(ss[1], ss[1])
            instrs.append(SubsortDecl(SubsortBody(IdList(qss0), IdList(qss1))))
        # declare translated ops
        for od in bp.op_decls:
            # obtain translated versions of domain sort names
            fromSortsRes = []
            for sn in od.from_sorts.ids:
                # sorts decl'd in imported modules are not in the dict
                qsn = d.get(sn, sn)
                fromSortsRes.append(qsn)
            # obtain translated versions of op ids
            opNamesRes = []
            for on in od.ops.ids:
                qon = d.get(on, on)
                opNamesRes.append(qon)
            # obtain translated version of range sort name
            toSortRes = d.get(od.to_sort, od.to_sort)
            # declare translated ops
            instrs.append(OpDecl(IdList(*opNamesRes), IdList(*fromSortsRes),
                                 od.arrow, toSortRes))
        # declare translated vars
        for vd in bp.var_decls:
            varNamesRes = []
            for vn in vd.ids.ids:
                qvn = d.get(vn, vn)
                varNamesRes.append(qvn)
            toSortRes = d.get(vd.sort, vd.sort)
            instrs.append(VarDecl(IdList(*varNamesRes), toSortRes))
        return ModBody(*instrs)

    @singledispatchmethod
    def __qualify(self, arg):
        raise NotImplementedError(
            f"There is no instantiation method for {type(arg)}.")

    @__qualify.register
    def _(self, ac: AtomicCondition) -> AtomicCondition:
        raise Exception("Method cannot be called for AtomicCondition,"
                        " but for some of its subclasses.")

    @__qualify.register
    def _(self, ec: EqualCondition) -> EqualCondition:
        return EqualCondition(Term(self.__qTermStr(str(ec.lhs))),
                              Term(self.__qTermStr(str(ec.rhs))))

    @__qualify.register
    def _(self, ac: AssignCondition) -> AssignCondition:
        return AssignCondition(Term(self.__qTermStr(str(ac.lhs))),
                               Term(self.__qTermStr(str(ac.rhs))))

    @__qualify.register
    def _(self, mc: MemberCondition) -> MemberCondition:
        return MemberCondition(Term(self.__qTermStr(str(mc.lhs))),
                               self.__qId(mc.rhs))

    @__qualify.register
    def _(self, rc: RewCondition) -> RewCondition:
        return RewCondition(Term(self.__qTermStr(str(rc.lhs))),
                            Term(self.__qTermStr(str(rc.rhs))))

    @__qualify.register
    def _(self, bc: BoolCondition) -> BoolCondition:
        return BoolCondition(Term(self.__qTermStr(str(bc.text))))

    @__qualify.register
    def _(self, c: Condition) -> Condition:
        return Condition(*[self.__qualify(ac) for ac in c.conds])

    @__qualify.register
    def _(self, il: IdList) -> IdList:
        return IdList(*[self.__qId(i) for i in il.ids])

    @__qualify.register
    def _(self, i: Instruction) -> Instruction:
        return i

    @__qualify.register
    def _(self, sd: SortDecl) -> SortDecl:
        return SortDecl(IdList(*[self.__qId(sn) for sn in sd.sorts.ids]))

    @__qualify.register
    def _(self, ssd: SubsortDecl) -> SubsortDecl:
        return SubsortDecl(SubsortBody(
            *[self.__qualify(il) for il in ssd.body.sort_lists]))

    @__qualify.register
    def _(self, od: OpDecl) -> OpDecl:
        return OpDecl(self.__qualify(od.ops),
                      self.__qualify(od.from_sorts),
                      od.arrow,
                      self.__qId(od.to_sort),
                      self.__qualify(od.attrs))

    @__qualify.register
    def _(self, al: AttrList) -> AttrList:
        res = []
        inId = False
        for a in al.attrs:
            if isinstance(a, AttrId):
                res.append(AttrId(self.__qId(a.term)))
            else:
                res.append(a)
        return AttrList(*res)

    @__qualify.register
    def _(self, vd: VarDecl) -> VarDecl:
        return VarDecl(self.__qualify(vd.ids), self.__qId(vd.sort))

    @__qualify.register
    def _(self, e: Eq) -> Eq:
        return Eq(Term(self.__qTermStr(str(e.lhs))),
                  Term(self.__qTermStr(str(e.rhs))), e.attrs)

    @__qualify.register
    def _(self, ce: CondEq) -> CondEq:
        return CondEq(Term(self.__qTermStr(str(ce.lhs))),
                      Term(self.__qTermStr(str(ce.rhs))),
                      self.__qualify(ce.cond), ce.attrs)

    @__qualify.register
    def _(self, m: Mb) -> Mb:
        return Mb(Term(self.__qTermStr(str(m.term))),
                  self.__qId(m.sort))

    @__qualify.register
    def _(self, cm: CondMb) -> CondMb:
        return CondMb(Term(self.__qTermStr(str(cm.term))),
                      self.__qId(cm.sort),
                      self.__qualify(cm.cond))

    @__qualify.register
    def _(self, r: Rule) -> Rule:
        return Rule(Term(self.__qTermStr(str(r.lhs))),
                    Term(self.__qTermStr(str(r.rhs))), r.attrs)

    @__qualify.register
    def _(self, cr: CondRule) -> CondRule:
        return CondRule(Term(self.__qTermStr(str(cr.lhs))),
                        Term(self.__qTermStr(str(cr.rhs))),
                        self.__qualify(cr.cond), cr.attrs)

    @__qualify.register
    def _(self, er: EgRule) -> EgRule:
        return EgRule(Term(self.__qTermStr(str(er.lhs))),
                      Term(self.__qTermStr(str(er.mid))),
                      Term(self.__qTermStr(str(er.rhs))), er.attrs)

    @__qualify.register
    def _(self, cer: CondEgRule) -> CondEgRule:
        return CondEgRule(Term(self.__qTermStr(str(cer.lhs))),
                          Term(self.__qTermStr(str(cer.mid))),
                          Term(self.__qTermStr(str(cer.rhs))),
                          self.__qualify(cer.cond), cer.attrs)

    def __qTermStr(self, termStr: str) -> str:
        term = self.__rrSignMod.parseTerm(termStr)
        qTerm = self.__qTerm(term)
        return qTerm.prettyPrint(maude.PRINT_MIXFIX | maude.PRINT_NUMBER)

    def __qId(self, name: str) -> str:
        """This method is to be used for sort ids, op ids, and var ids,
        whenever they are used as identifiers. For ops and vars this is when
        they are declared, but not when they are a term or part of a term. For
        sort ids, however, this is to be used at all times: declarations,
        membership axioms..."""
        return self.__qDict.get(name, name)

    def __qTerm(self, term: maude.Term) -> maude.Term:
        if self.__isVarTerm(term):
            return self.__qVarTerm(term)
        elif Qualifier.__isQidTerm(term):
            return self.__qQidTerm(term)
        elif Qualifier.__isNatTerm(term):
            return self.__qNatTerm(term)
        else:
            qArgs = []
            for a in term.arguments():
                qArgs.append(self.__qTerm(a))
            qOp = self.__qSymbol(term.symbol())
            return qOp.makeTerm(qArgs)

    def __isVarTerm(self, term: maude.Term) -> bool:
        # `term` must be a term in `self.__fromMod`
        return term.symbol() == self.__rrSignMod.parseTerm(
            f'<FfAaKkEe|VvAaRr>:{term.getSort()}').symbol()

    @staticmethod
    def __isQidTerm(term: maude.Term) -> bool:
        # `term` must be a term in `self.__fromMod`
        return str(term.symbol()) == '<Qids>'

    @staticmethod
    def __isNatTerm(term: maude.Term) -> bool:
        # `term` must be a term in `self.__fromMod`
        return str(term.symbol()) == 's_'

    def __qVarTerm(self, varTerm: maude.Term) -> maude.Term:
        """`varTerm` is assumed to be a valid variable term (either in the form
        `X` or `X:S`) in `self.__rrSignMod`. There are three cases: the
        variable appears as `X`, and then it was declared; it appears as `X:S`
        and was declared, and the meaning it the same as in the previous case;
        or it appears as `Y:S` and was not declared. In all cases, the name is
        qualified and the sort is or is not according to the dictionary."""
        varStr = str(varTerm)
        sortName = str(varTerm.getSort())
        if varStr.endswith(':' + sortName):
            varName = varStr.rsplit(':', 1)[0]
            if varName in self.__qDict.keys():
                newVarName = self.__qDict[varName]
            else:
                newVarName = buildName(self.__q, varName)
            newSortName = self.__qDict.get(sortName, sortName)
            res = self.__rrQSignMod.parseTerm(newVarName + ':' + newSortName)
        else:
            # If a var appears as `X` in a term, it must have been declared,
            # and so it must be in the dictionary.
            res = self.__rrQSignMod.parseTerm(self.__qDict[varStr])
        return res

    def __qQidTerm(self, qidTerm: maude.Term) -> maude.Term:
        """`idTerm` is assumed to be a valid Qid. It is not really translated,
        but a reproduction of it in `toModStr` is produced."""
        return self.__rrQSignMod.parseTerm(str(qidTerm))

    def __qNatTerm(self, natTerm: maude.Term) -> maude.Term:
        """`idTerm` is assumed to be a valid Nat. It is not really translated,
        but a reproduction of it in `toModStr` is produced."""
        return self.__rrQSignMod.parseTerm(str(natTerm))

    def __qSymbol(self, op: maude.Symbol) -> maude.Symbol:
        """Returns the translated version of the symbol `op`, according to the
        dictionary. If its name is not in the dictionary, the only cause can
        be that it is declared in an imported module, so the translated version
        must have the same name as the original."""
        qDomainKindList = []
        for i in range(op.arity()):
            qDomainKindList.append(self.__qKind(op.domainKind(i)))
        qRangeKind = self.__qKind(op.getRangeSort().kind())
        qName = self.__qDict.get(str(op), str(op))
        res = self.__rrQSignMod.findSymbol(qName, qDomainKindList, qRangeKind)
        return res

    def __qSort(self, sort: maude.Sort) -> maude.Sort:
        """If `sort`'s name is in the dictionary, the sort for its translation
        is returned. Otherwise, the only reason can be that it is declared in
        an imported module, so the sort with the same name is returned."""
        sortName = str(sort)
        trSortName = self.__qDict.get(sortName, sortName)
        return self.__rrQSignMod.findSort(trSortName)

    def __qKind(self, kind: maude.Kind) -> maude.Kind:
        return self.__qSort(kind.sort(1)).kind()
