#!/usr/bin/env python3.9


from importlib import resources
import re
import pytest
import lark
import terminals
from other import astToObjModel
from instantiator import Instantiator
from tools import larkify


def simple(s: str) -> str:
    s1 = s.replace('\t', ' ')
    s2 = s1.replace('\n', ' ')
    s3 = re.sub(' +', ' ', s2)
    s4 = s3.strip()
    return s4


@pytest.fixture
def inr() -> Instantiator:
    return Instantiator({"M1": "M2", "N1": "N2"},
                        {"M1$p1": "M2$p2", "N1$q1": "N2$q2"})


@pytest.fixture
def gr() -> str:
    larkText = resources.read_text('grammars', 'cmaude.lark')
    cMaudeGrammarText = larkify(terminals.terminal_defs() + larkText)
    return cMaudeGrammarText


# noinspection PyUnresolvedReferences
def test__syncon(inr, gr):
    orig = "sync M1 || N1 on M1$p1 = N1$q1 ."
    res = "sync M2 || N2 on M2$p2 = N2$q2 ."
    ast = lark.Lark(gr, start="sync_on").parse(orig)
    om = astToObjModel(ast)
    iom = inr._Instantiator__instantiate(om)
    assert simple(str(iom)) == res


# noinspection PyUnresolvedReferences
def test__eq(inr, gr):
    orig = "ceq M1$p1 = b if uno = dos ."
    res = "ceq M2$p2 = b if uno = dos ."
    ast = lark.Lark(gr, start="cond_eq").parse(orig)
    om = astToObjModel(ast)
    iom = inr._Instantiator__instantiate(om)
    assert simple(str(iom)) == res


# noinspection PyUnresolvedReferences
def test__inh(inr, gr):
    orig = "inh p = N1$q1 ."
    res = "inh p = N2$q2 ."
    ast = lark.Lark(gr, start="inh_stat").parse(orig)
    om = astToObjModel(ast)
    iom = inr._Instantiator__instantiate(om)
    assert simple(str(iom)) == res


def test__top(gr):
    p = lark.Lark(gr, start="maude_top")
    orig = """
        th T1 is
            pr NAT .
            ppt p1 : -> Nat .
        endth
        th T2 is
            pr NAT .
            ppt p2 : -> Nat .
        endth
        aemod M1 is
            pr NAT .
            ppt q1 : -> Nat .
        endaem
        aemod M2 is
            pr NAT .
            ppt p2 : -> Nat .
        endaem
        view V1 from T1 to M1 is
            ppt p1 to q1 .
        endv
        view V2 from T2 to M2 is
        endv
        emod M{X :: T1, Y :: T2} is
            sync X || Y
                on X$p1 = Y$p2 .
        endem
        emod M12 is
            inc M{V1, V2} .
        endem
        """
    res = """
        th T1 is
            pr NAT .
            ppt p1 : -> Nat .
        endth
        th T2 is
            pr NAT .
            ppt p2 : -> Nat .
        endth
        aemod M1 is
            pr NAT .
            ppt q1 : -> Nat .
        endaem
        aemod M2 is
            pr NAT .
            ppt p2 : -> Nat .
        endaem
        view V1 from T1 to M1 is
            ppt p1 to q1 .
        endv
        view V2 from T2 to M2 is
        endv
        emod <M|V1|V2> is
            sync M1 || M2
                on M1$q1 = M2$p2 .
        endem
        emod M12 is
            inc <M|V1|V2> .
        endem
        """
    ast = p.parse(orig)
    top = astToObjModel(ast)
    newTop = Instantiator.params_to_instances(top)
    assert simple(str(newTop)) == simple(res)
