#!/usr/bin/env python3.9


from importlib import resources
import pytest
import lark
from other import astToObjModel, get_mod_by_name
from objmodel import MaudeTop
from qualifier import Qualifier
import terminals
from tools import larkify


@pytest.fixture
def qr() -> Qualifier:
    topStr =\
        '''
        fmod KK is
            pr QID .
            sorts S T .
            op s : -> S .
            op f : S -> S .
            op __ : S S -> S .
            var X : S .
            op g : S -> Bool .
            op h : Bool -> Bool .
            var B : Bool .
        endfm
        '''
    larkText = resources.read_text('grammars', 'cmaude.lark')
    cMaudeGrammarText = terminals.terminal_defs() + larkText
    cMaudeGrammar = lark.Lark(larkify(cMaudeGrammarText), start="maude_top",
                              maybe_placeholders=True)
    topAst = cMaudeGrammar.parse(topStr)
    top: MaudeTop = astToObjModel(topAst)
    b = get_mod_by_name(top, 'KK').body
    return Qualifier(b, 'KK', top)


# noinspection PyUnresolvedReferences
def test__isVarTerm(qr):
    t = qr._Qualifier__rrSignMod.parseTerm('X')
    assert qr._Qualifier__isVarTerm(t)
    t = qr._Qualifier__rrSignMod.parseTerm('Y:S')
    assert qr._Qualifier__isVarTerm(t)
    t = qr._Qualifier__rrSignMod.parseTerm('s')
    assert not qr._Qualifier__isVarTerm(t)


# noinspection PyUnresolvedReferences
def test__qVarTerm(qr):
    t = qr._Qualifier__rrSignMod.parseTerm('X')
    tt = qr._Qualifier__rrQSignMod.parseTerm('<KK|X>')
    assert str(qr._Qualifier__qVarTerm(t)) == str(tt)

    t = qr._Qualifier__rrSignMod.parseTerm('X:S')
    tt = qr._Qualifier__rrQSignMod.parseTerm('<KK|X>:<KK|S>')
    assert str(qr._Qualifier__qVarTerm(t)) == str(tt)

    t = qr._Qualifier__rrSignMod.parseTerm('Y:S')
    tt = qr._Qualifier__rrQSignMod.parseTerm('<KK|Y>:<KK|S>')
    assert str(qr._Qualifier__qVarTerm(t)) == str(tt)


# noinspection PyUnresolvedReferences
def test__isQidTerm(qr):
    t = qr._Qualifier__rrSignMod.parseTerm("'Q.Qid")
    assert qr._Qualifier__isQidTerm(t)
    t = qr._Qualifier__rrSignMod.parseTerm("'Q")
    assert qr._Qualifier__isQidTerm(t)
    t = qr._Qualifier__rrSignMod.parseTerm("B:Bool")
    assert not qr._Qualifier__isQidTerm(t)
    t = qr._Qualifier__rrSignMod.parseTerm("X")
    assert not qr._Qualifier__isQidTerm(t)
    t = qr._Qualifier__rrSignMod.parseTerm('f(s) X')
    assert not qr._Qualifier__isQidTerm(t)


# noinspection PyUnresolvedReferences
def test__qQidTerm(qr):
    t = qr._Qualifier__rrSignMod.parseTerm("'Q.Qid")
    tt = qr._Qualifier__rrQSignMod.parseTerm("'Q.Qid")
    assert str(qr._Qualifier__qQidTerm(t)) == str(tt)


# noinspection PyUnresolvedReferences
def test__qSort(qr):
    s = qr._Qualifier__rrSignMod.findSort('S')
    ts = qr._Qualifier__rrQSignMod.findSort('<KK|S>')
    assert qr._Qualifier__qSort(s).equal(ts)
    s = qr._Qualifier__rrSignMod.findSort('Bool')
    ts = qr._Qualifier__rrQSignMod.findSort('Bool')
    assert qr._Qualifier__qSort(s).equal(ts)


# noinspection PyUnresolvedReferences
def test__qKind(qr):
    k = qr._Qualifier__rrSignMod.findSort('S').kind()
    tk = qr._Qualifier__rrQSignMod.findSort('<KK|S>').kind()
    assert qr._Qualifier__qKind(k).equal(tk)
    k = qr._Qualifier__rrSignMod.findSort('Bool').kind()
    tk = qr._Qualifier__rrQSignMod.findSort('Bool').kind()
    assert qr._Qualifier__qKind(k).equal(tk)


# noinspection PyUnresolvedReferences
def test__trSymbol(qr):
    s = qr._Qualifier__rrSignMod.parseTerm('s').symbol()
    ts = qr._Qualifier__rrQSignMod.parseTerm('<KK|s>').symbol()
    assert qr._Qualifier__qSymbol(s).equal(ts)

    k = qr._Qualifier__rrSignMod.findSort('S').kind()
    tk = qr._Qualifier__rrQSignMod.findSort('<KK|S>').kind()
    s = qr._Qualifier__rrSignMod.findSymbol('__', [k, k], k)
    ts = qr._Qualifier__rrQSignMod.findSymbol('<KK|__>', [tk, tk], tk)
    assert qr._Qualifier__qSymbol(s).equal(ts)

    k = qr._Qualifier__rrSignMod.findSort('S').kind()
    tk = qr._Qualifier__rrQSignMod.findSort('<KK|S>').kind()
    k2 = qr._Qualifier__rrSignMod.findSort('Bool').kind()
    tk2 = qr._Qualifier__rrQSignMod.findSort('Bool').kind()
    s = qr._Qualifier__rrSignMod.findSymbol('g', [k], k2)
    ts = qr._Qualifier__rrQSignMod.findSymbol('<KK|g>', [tk], tk2)
    assert qr._Qualifier__qSymbol(s).equal(ts)

    k = qr._Qualifier__rrSignMod.findSort('Bool').kind()
    tk = qr._Qualifier__rrQSignMod.findSort('Bool').kind()
    s = qr._Qualifier__rrSignMod.findSymbol('h', [k], k)
    ts = qr._Qualifier__rrQSignMod.findSymbol('<KK|h>', [tk], tk)
    assert qr._Qualifier__qSymbol(s).equal(ts)


# noinspection PyUnresolvedReferences
def test__qTerm(qr):
    t = qr._Qualifier__rrSignMod.parseTerm('f(s) X')
    tt = qr._Qualifier__rrQSignMod.parseTerm('<KK| <KK|f>(<KK|s>) <KK|X> >')
    assert str(qr._Qualifier__qTerm(t)) == str(tt)

    t = qr._Qualifier__rrSignMod.parseTerm('h(B:Bool)')
    tt = qr._Qualifier__rrQSignMod.parseTerm('<KK|h>(<KK|B>:Bool)')
    assert str(qr._Qualifier__qTerm(t)) == str(tt)


# noinspection PyUnresolvedReferences
def test_qTermStr(qr):
    assert qr._Qualifier__qTermStr('1') == '1'
    assert qr._Qualifier__qTermStr('2') == '2'
