#!/usr/bin/env python3.9

from importlib import resources
import pytest
import re
from typing import List, Tuple
import lark
import terminals
from tools import larkify


def simple(s: str) -> str:
    s1 = s.replace('\t', ' ')
    s2 = re.sub(' +', ' ', s1)
    s3 = s2.strip()
    return s3


@pytest.fixture
def listAndTree() -> Tuple[List[str], lark.Tree]:
    larkText = resources.read_text('grammars', 'cmaude.lark')
    cMaudeGrammarText = terminals.terminal_defs() + larkText
    p = lark.Lark(larkify(cMaudeGrammarText), start="mod_body")
    text =\
        '''eq lhs = rhs .
eq lhs1 lhs2 = rhs1 rhs2 rhs3 .
eq   lhs1   lhs2   =   rhs1   rhs2   rhs3   [owise]   .
eq lhs1 lhs2 = (rhs1 rhs2) rhs3 [owise comm] .
eq lhs1 lhs2 = if a then b else c fi .
ceq lhs1 lhs2 = rhs1 rhs2 rhs3 if a = b .
ceq lhs1 lhs2 = rhs1 rhs2 rhs3 if a := b .
ceq lhs1 lhs2 = rhs1 rhs2 rhs3 if a : S .
ceq lhs1 lhs2 = rhs1 rhs2 rhs3 if a = b [owise] .
ceq lhs1 lhs2 = (if a then b else c fi) if a = b [owise] .
eq lhs1 lhs2 = rhs1 [rhs2] [rhs3] [owise] .
eq lhs1 lhs2 = rhs1 ([rhs2]) ([rhs3] [owise]) .
rl lhs => rhs .
rl lhs1 lhs2 => if a then b else c fi .
crl lhs1 lhs2 => rhs1 rhs2 rhs3 if a => b .
crl lhs1 lhs2 => rhs1 rhs2 rhs3 if a = b [owise] .
crl lhs1 lhs2 => (if a then (b) else c fi) if a = b [owise] .
crl lhs =[ mid ]=> rhs if a = b [owise] .
mb term : Sort .
cmb term : Sort if cl = cr .
eq a b []={} rhs .
rl lhs =[ (if a then b else c fi) mid ]=> rhs .
ceq a = b if cond .
crl a => b if cond .
ceq a = b if cond [owise] .
eq a = b [abc] .
eq lhs = rhs [id: abc] .
eq lhs = rhs [comm id: abc] .'''
    t = p.parse(text)
    return text.split('\n'), t


# def testKK():
#     # These can be used to make pytest print the result in the lhs.
#     assert bnf.token('[', '(', 'abc') == ''
#     assert bnf.anyTokenBut('=') == ''


def test_i00(listAndTree):
    instr = 'eq lhs = rhs .'
    i = listAndTree[0].index(instr)
    t = listAndTree[1].children[i]
    assert simple(t.pretty()) == simple('''
eq
    term___eq_lhs    lhs
    term___rhs       rhs
''')


def test_i01(listAndTree):
    instr = 'eq lhs1 lhs2 = rhs1 rhs2 rhs3 .'
    i = listAndTree[0].index(instr)
    t = listAndTree[1].children[i]
    assert simple(t.pretty()) == simple('''
eq
    term___eq_lhs
        lhs1
        lhs2
    term___rhs
        rhs1
        rhs2
        rhs3
''')


def test_i02(listAndTree):
    instr = 'eq   lhs1   lhs2   =   rhs1   rhs2   rhs3   [owise]   .'
    i = listAndTree[0].index(instr)
    t = listAndTree[1].children[i]
    assert simple(t.pretty()) == simple('''
eq
    term___eq_lhs
        lhs1
        lhs2
    term___rhs
        rhs1
        rhs2
        rhs3
    attr_list
        attr_owise
''')


def test_i03(listAndTree):
    instr = 'eq lhs1 lhs2 = (rhs1 rhs2) rhs3 [owise comm] .'
    i = listAndTree[0].index(instr)
    t = listAndTree[1].children[i]
    assert simple(t.pretty()) == simple('''
eq
    term___eq_lhs
        lhs1
        lhs2
    term___rhs
        (rhs1 rhs2)
        rhs3
    attr_list
        attr_owise
        attr_comm
''')


def test_i04(listAndTree):
    instr = 'eq lhs1 lhs2 = if a then b else c fi .'
    i = listAndTree[0].index(instr)
    t = listAndTree[1].children[i]
    assert simple(t.pretty()) == simple('''
eq
    term___eq_lhs
        lhs1
        lhs2
    term___rhs
        if
        a
        then
        b
        else
        c
        fi
''')


def test_i05(listAndTree):
    instr = 'ceq lhs1 lhs2 = rhs1 rhs2 rhs3 if a = b .'
    i = listAndTree[0].index(instr)
    t = listAndTree[1].children[i]
    assert simple(t.pretty()) == simple('''
cond_eq
    term___eq_lhs
        lhs1
        lhs2
    term___ceq_crl_rhs
        rhs1
        rhs2
        rhs3
    condition
        equal_condition
            term___cond_lhs    a
            term___rhs         b
''')


def test_i06(listAndTree):
    instr = 'ceq lhs1 lhs2 = rhs1 rhs2 rhs3 if a := b .'
    i = listAndTree[0].index(instr)
    t = listAndTree[1].children[i]
    assert simple(t.pretty()) == simple('''
cond_eq
    term___eq_lhs
        lhs1
        lhs2
    term___ceq_crl_rhs
        rhs1
        rhs2
        rhs3
    condition
        assign_condition
            term___cond_lhs    a
            term___rhs         b
''')


def test_i07(listAndTree):
    instr = 'ceq lhs1 lhs2 = rhs1 rhs2 rhs3 if a : S .'
    i = listAndTree[0].index(instr)
    t = listAndTree[1].children[i]
    assert simple(t.pretty()) == simple('''
cond_eq
    term___eq_lhs
        lhs1
        lhs2
    term___ceq_crl_rhs
        rhs1
        rhs2
        rhs3
    condition
        member_condition
            term___cond_lhs    a
            S
''')


def test_i08(listAndTree):
    instr = 'ceq lhs1 lhs2 = rhs1 rhs2 rhs3 if a = b [owise] .'
    i = listAndTree[0].index(instr)
    t = listAndTree[1].children[i]
    assert simple(t.pretty()) == simple('''
cond_eq
    term___eq_lhs
        lhs1
        lhs2
    term___ceq_crl_rhs
        rhs1
        rhs2
        rhs3
    condition
        equal_condition
            term___cond_lhs    a
            term___rhs         b
    attr_list
        attr_owise
''')


def test_i09(listAndTree):
    instr = 'ceq lhs1 lhs2 = (if a then b else c fi) if a = b [owise] .'
    i = listAndTree[0].index(instr)
    t = listAndTree[1].children[i]
    assert simple(t.pretty()) == simple('''
cond_eq
    term___eq_lhs
        lhs1
        lhs2
    term___ceq_crl_rhs    (if a then b else c fi)
    condition
        equal_condition
            term___cond_lhs    a
            term___rhs         b
    attr_list
        attr_owise
''')


def test_i10(listAndTree):
    instr = 'eq lhs1 lhs2 = rhs1 [rhs2] [rhs3] [owise] .'
    i = listAndTree[0].index(instr)
    t = listAndTree[1].children[i]
    assert simple(t.pretty()) == simple('''
eq
    term___eq_lhs
        lhs1
        lhs2
    term___rhs
        rhs1
        [
        rhs2
        ]
        [
        rhs3
        ]
    attr_list
        attr_owise
''')


def test_i11(listAndTree):
    instr = 'eq lhs1 lhs2 = rhs1 ([rhs2]) ([rhs3] [owise]) .'
    i = listAndTree[0].index(instr)
    t = listAndTree[1].children[i]
    assert simple(t.pretty()) == simple('''
eq
    term___eq_lhs
        lhs1
        lhs2
    term___rhs
        rhs1
        ([rhs2])
        ([rhs3] [owise])
''')


def test_i12(listAndTree):
    instr = 'rl lhs => rhs .'
    i = listAndTree[0].index(instr)
    t = listAndTree[1].children[i]
    assert simple(t.pretty()) == simple('''
rule
    term___rl_lhs    lhs
    term___rhs       rhs
''')


def test_i13(listAndTree):
    instr = 'rl lhs1 lhs2 => if a then b else c fi .'
    i = listAndTree[0].index(instr)
    t = listAndTree[1].children[i]
    assert simple(t.pretty()) == simple('''
rule
    term___rl_lhs
        lhs1
        lhs2
    term___rhs
        if
        a
        then
        b
        else
        c
        fi
''')


def test_i14(listAndTree):
    instr = 'crl lhs1 lhs2 => rhs1 rhs2 rhs3 if a => b .'
    i = listAndTree[0].index(instr)
    t = listAndTree[1].children[i]
    assert simple(t.pretty()) == simple('''
cond_rule
    term___rl_lhs
        lhs1
        lhs2
    term___ceq_crl_rhs
        rhs1
        rhs2
        rhs3
    condition
        rew_condition
            term___cond_lhs    a
            term___rhs         b
''')


def test_i15(listAndTree):
    instr = 'crl lhs1 lhs2 => rhs1 rhs2 rhs3 if a = b [owise] .'
    i = listAndTree[0].index(instr)
    t = listAndTree[1].children[i]
    assert simple(t.pretty()) == simple('''
cond_rule
    term___rl_lhs
        lhs1
        lhs2
    term___ceq_crl_rhs
        rhs1
        rhs2
        rhs3
    condition
        equal_condition
            term___cond_lhs    a
            term___rhs         b
    attr_list
        attr_owise
''')


def test_i16(listAndTree):
    instr = 'crl lhs1 lhs2 => (if a then (b) else c fi) if a = b [owise] .'
    i = listAndTree[0].index(instr)
    t = listAndTree[1].children[i]
    assert simple(t.pretty()) == simple('''
cond_rule
    term___rl_lhs
        lhs1
        lhs2
    term___ceq_crl_rhs    (if a then (b) else c fi)
    condition
        equal_condition
            term___cond_lhs    a
            term___rhs         b
    attr_list
        attr_owise
''')


def test_i17(listAndTree):
    instr = 'crl lhs =[ mid ]=> rhs if a = b [owise] .'
    i = listAndTree[0].index(instr)
    t = listAndTree[1].children[i]
    assert simple(t.pretty()) == simple('''
cond_eg_rule
    term___rl_lhs    lhs
    term___erl_mid    mid
    term___ceq_crl_rhs    rhs
    condition
        equal_condition
            term___cond_lhs    a
            term___rhs         b
    attr_list
        attr_owise
''')


def test_i18(listAndTree):
    instr = 'mb term : Sort .'
    i = listAndTree[0].index(instr)
    t = listAndTree[1].children[i]
    assert simple(t.pretty()) == simple('''
mb
    term___mb_lhs    term
    Sort
''')


def test_i19(listAndTree):
    instr = 'cmb term : Sort if cl = cr .'
    i = listAndTree[0].index(instr)
    t = listAndTree[1].children[i]
    assert simple(t.pretty()) == simple('''
cond_mb
    term___mb_lhs    term
    Sort
    condition
        equal_condition
            term___cond_lhs    cl
            term___rhs    cr
''')


def test_i20(listAndTree):
    instr = 'eq a b []={} rhs .'
    i = listAndTree[0].index(instr)
    t = listAndTree[1].children[i]
    assert simple(t.pretty()) == simple('''
eq
    term___eq_lhs
        a
        b
        [
        ]
    term___rhs
        {
        }
        rhs
''')


def test_i21(listAndTree):
    instr = 'rl lhs =[ (if a then b else c fi) mid ]=> rhs .'
    i = listAndTree[0].index(instr)
    t = listAndTree[1].children[i]
    assert simple(t.pretty()) == simple('''
eg_rule
    term___rl_lhs    lhs
    term___erl_mid
        (if a then b else c fi)
        mid
    term___rhs    rhs
''')


def test_i22(listAndTree):
    instr = 'ceq a = b if cond .'
    i = listAndTree[0].index(instr)
    t = listAndTree[1].children[i]
    assert simple(t.pretty()) == simple('''
cond_eq
    term___eq_lhs    a
    term___ceq_crl_rhs    b
    condition
        bool_condition
            term___rhs    cond
''')


def test_i23(listAndTree):
    instr = 'crl a => b if cond .'
    i = listAndTree[0].index(instr)
    t = listAndTree[1].children[i]
    assert simple(t.pretty()) == simple('''
cond_rule
    term___rl_lhs    a
    term___ceq_crl_rhs    b
    condition
        bool_condition
            term___rhs    cond
''')


def test_i24(listAndTree):
    instr = 'ceq a = b if cond [owise] .'
    i = listAndTree[0].index(instr)
    t = listAndTree[1].children[i]
    assert simple(t.pretty()) == simple('''
cond_eq
    term___eq_lhs    a
    term___ceq_crl_rhs    b
    condition
        bool_condition
            term___rhs    cond
    attr_list
        attr_owise
''')


def test_i25(listAndTree):
    instr = 'eq a = b [abc] .'
    i = listAndTree[0].index(instr)
    t = listAndTree[1].children[i]
    assert simple(t.pretty()) == simple('''
eq
    term___eq_lhs    a
    term___rhs
        b
        [
        abc
        ]
''')


def test_i26(listAndTree):
    instr = 'eq lhs = rhs [id: abc] .'
    i = listAndTree[0].index(instr)
    t = listAndTree[1].children[i]
    assert simple(t.pretty()) == simple('''
eq
    term___eq_lhs    lhs
    term___rhs       rhs
    attr_list
        attr_id    abc
''')


def test_i27(listAndTree):
    instr = 'eq lhs = rhs [comm id: abc] .'
    i = listAndTree[0].index(instr)
    t = listAndTree[1].children[i]
    assert simple(t.pretty()) == simple('''
eq
    term___eq_lhs    lhs
    term___rhs       rhs
    attr_list
        attr_comm
        attr_id    abc
''')
