#!/usr/bin/env python3.9


from typing import Tuple
from objmodel import *


def get_mod_by_name(top: MaudeTop, name: str) -> Module:
    for e in top.elems:
        if isinstance(e, Module):
            if e.name == name:
                return e


def get_view_by_name(top: MaudeTop, name: str) -> View:
    for e in top.elems:
        if isinstance(e, View):
            if e.name == name:
                return e


def get_theory_by_name(top: MaudeTop, name: str) -> Theory:
    for e in top.elems:
        if isinstance(e, Theory):
            if e.name == name:
                return e


def get_ppt_mappings(top: MaudeTop, v: View) -> List[Tuple[str, str]]:
    # Grab list of ppt mappings from view.
    res = []
    for i in v.body.maps:
        if isinstance(i, PptMapping):
            res += [(i.lhs, i.rhs)]
    # Grab list of ppts from theory, then add missing default mappings.
    thPptList = []
    for i in sys_flat_theory(top,
                             get_theory_by_name(top, v.mod_from)).body.instrs:
        if isinstance(i, PptDecl):
            for x in i.ppts.ids:
                thPptList.append(x)
    res += [(x, x) for x in thPptList if x not in [a for (a, b) in res]]
    return res


def sys_flat_theory(top: MaudeTop, t: Theory) -> Theory:
    instrs = []
    for i in t.body.instrs:
        if isinstance(i, Protecting) \
                or isinstance(i, Extending) \
                or isinstance(i, Including):
            t2 = get_theory_by_name(top, i.mod_expr.name)
            if isinstance(t2, Theory):
                instrs += sys_flat_theory(top, t2).body.instrs
            else:
                instrs.append(i)
        else:
            instrs.append(i)
    return t.__class__(t.name, ThBody(*instrs))


def get_context_str(top: MaudeTop, instrs: List[Instruction], done: List[str])\
        -> str:
    """It takes a list of instructions, of which only the importing ones
    are relevant. It looks for all modules stored in `top`
    that are required for `instrs` to make sense, that is, it
    looks recursively for all *non-flattable* modules imported. The
    parameter `done` is used to keep track of modules already taken care
    of, in case they appear again. The *non-flattable* part is important.
    The thing is that flattable all_imports are assumed to have been already
    flattened before calling to this method. Non-flattable all_imports are
    those functional modules that are imported in `extending` or
    `protecting` modes. These are the only ones this method care about.
    The result is given as a (possibly long) string with the text of all
    such modules."""
    res = ""
    for i in instrs:
        if isinstance(i, Protecting) or isinstance(i, Extending):
            mn = i.mod_expr.name
            if mn not in done:
                m = get_mod_by_name(top, mn)
                if isinstance(m, FModule):
                    res += get_context_str(top, m.body.instrs, done)
                    res += str(m) + "\n"
                    done.append(mn)
    return res
